#!/usr/bin/env python
# flake8: noqa: E402
import argparse
import logging
import sys

from pathlib import Path


"""
Update a database based on a RPSL file.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.rpki.validators import BulkRouteROAValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.mirroring.parsers import MirrorUpdateFileImportParser
from irrd.conf import config_init, CONFIG_PATH_DEFAULT, get_setting

def update(source, filename) -> int:
    if any([
        get_setting(f'sources.{source}.import_source'),
        get_setting(f'sources.{source}.import_serial_source')
    ]):
        print(f'Error: to use this command, import_source and import_serial_source '
              f'for source {source} must not be set.')
        return 2

    dh = DatabaseHandler()
    roa_validator = BulkRouteROAValidator(dh)
    parser = MirrorUpdateFileImportParser(
        source, filename, database_handler=dh,
        direct_error_return=True, roa_validator=roa_validator)
    error = parser.run_import()
    if error:
        dh.rollback()
    else:
        dh.commit()
    dh.close()
    if error:
        print(f'Error occurred while processing object:\n{error}')
        return 1
    return 0


def main():  # pragma: no cover
    description = """Update a database based on a RPSL file."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('--source', dest='source', type=str, required=True,
                        help=f'name of the source, e.g. NTTCOM')
    parser.add_argument('input_file', type=str,
                        help='the name of a file to read')
    args = parser.parse_args()

    config_init(args.config_file_path)
    if get_setting('database_readonly'):
        print('Unable to run, because database_readonly is set')
        sys.exit(-1)

    sys.exit(update(args.source, args.input_file))


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python
# flake8: noqa: E402

"""
Submit a raw update message, i.e. without email headers.
This supports password/override attributes, but is not
compatible with PGP signatures.

The message is always read from stdin.

Prints a report of the results, which would otherwise
be sent to a user by e-mail.
"""
import argparse
import sys

from pathlib import Path

sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import config_init, CONFIG_PATH_DEFAULT
from irrd.updates.handler import ChangeSubmissionHandler


def main(data):
    handler = ChangeSubmissionHandler(data)
    print(handler.submitter_report())


if __name__ == '__main__':  # pragma: no cover
    description = """Process a raw update message, i.e. without email headers. Authentication is still checked, 
                     but PGP is not supported. Message is always read from stdin, and a report is printed to stdout."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    args = parser.parse_args()

    config_init(args.config_file_path)

    main(sys.stdin.read())



import datrie
from collections import defaultdict

import codecs
import socket
from IPy import IP
from typing import Optional, List, Tuple, Dict

from irrd.conf import RPKI_IRR_PSEUDO_SOURCE, get_setting
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery, ROADatabaseObjectQuery
from .importer import ROA
from .status import RPKIStatus

# Pregenerated conversion from binary to binary strings for performance
BYTE_BIN = [bin(byte)[2:].zfill(8) for byte in range(256)]

decode_hex = codecs.getdecoder("hex_codec")


class BulkRouteROAValidator:
    """
    The bulk route validator is optimised to validate large amounts
    of routes, e.g. all RPSL route(6)s in the DB.

    As a typical database includes several million route objects,
    this class builds a trie using datrie. All ROAs are then inserted
    into that tree. For each route to be validated, the tree is searched
    to find any matching ROAs.

    ROAs are inserted under a key that is a binary string representation of
    their prefix, including only network bits. For 192.0.2.0/24, the binary
    representation is 11000000000000000000001000000000, reducing this to
    network bits mean the key is 110000000000000000000010.
    As multiple ROAs may exist for the same prefix, the data under the
    110000000000000000000010 key is actually a list of tuples, each
    tuple representing a ROA.

    When searching for ROAs for a route, the entire network address
    is converted to binary, and also trimmed to network bits.
    E.g. for 192.0.2.0/25, this would be 1100000000000000000000100.
    The trie is then searched for any items that are a prefix (inclusive) of
    1100000000000000000000100, which includes 110000000000000000000010,
    the key for 192.0.2(.0). A ROA for 192.0.2.0/26 would have key
    11000000000000000000001000, and therefore will (and should)
    not be included in the validation process.
    """

    def __init__(self, dh: DatabaseHandler, roas: Optional[List[ROA]] = None):
        """
        Create a validator object. Can use either a list of ROA objects,
        or if not given, generates this from the database.
        Due to the overhead in preloading all ROAs, this is only effective
        when many routes have to be validated, otherwise it's more
        efficient to use SingleRouteROAValidator. The break even
        point is in the order of magnitude of checking 10.000 routes.
        """
        self.database_handler = dh

        self.excluded_sources = []
        for source, settings in get_setting('sources', {}).items():
            if settings.get('rpki_excluded'):
                self.excluded_sources.append(source)

        self.roa_tree4 = datrie.Trie('01')
        self.roa_tree6 = datrie.Trie('01')
        if roas is None:
            self._build_roa_tree_from_db()
        else:
            self._build_roa_tree_from_roa_objs(roas)

    def validate_all_routes(self, sources: List[str]=None) -> \
            Tuple[List[Dict[str, str]], List[Dict[str, str]], List[Dict[str, str]]]:
        """
        Validate all RPSL route/route6 objects.

        Retrieves all routes from the DB, and aggregates the validation results.
        Returns a tuple of three sets of RPSL route(6)'s:
        - one with routes that should be set to status VALID, but are not now
        - one with routes that should be set to status INVALID, but are not now
        - one with routes that should be set to status UNKNOWN, but are not now
        Each route is recorded as a dict, which has the fields shown
        in "columns" below.

        Routes where their current validation status in the DB matches the new
        validation result, are not included in the return value.
        """
        columns = ['pk', 'rpsl_pk', 'ip_first', 'prefix_length', 'asn_first', 'source',
                   'object_class', 'object_text', 'rpki_status', 'scopefilter_status']
        q = RPSLDatabaseQuery(column_names=columns, enable_ordering=False)
        q = q.object_classes(['route', 'route6'])
        if sources:
            q = q.sources(sources)
        routes = self.database_handler.execute_query(q)

        objs_changed: Dict[RPKIStatus, List[Dict[str, str]]] = defaultdict(list)

        for result in routes:
            # RPKI_IRR_PSEUDO_SOURCE objects are ROAs, and don't need validation.
            if result['source'] == RPKI_IRR_PSEUDO_SOURCE:
                continue

            current_status = result['rpki_status']
            result['old_status'] = current_status
            new_status = self.validate_route(result['ip_first'], result['prefix_length'],
                                             result['asn_first'], result['source'])
            if new_status != current_status:
                result['rpki_status'] = new_status
                objs_changed[new_status].append(result)

        return objs_changed[RPKIStatus.valid], objs_changed[RPKIStatus.invalid], objs_changed[RPKIStatus.not_found]

    def validate_route(self, prefix_ip: str, prefix_length: int, prefix_asn: int, source: str) -> RPKIStatus:
        """
        Validate a single route.

        A route is valid when at least one ROA is found that covers the prefix,
        with the same origin AS and a match on the max length in the ROA.
        A route is invalid when at least one ROA is found that covers the prefix,
        but none of the covering ROAs matched on both origin AS and max length.
        A route is not_found if no ROAs were found covering the prefix.
        """
        if source in self.excluded_sources:
            return RPKIStatus.not_found

        ip_version, ip_bin_str = self._ip_to_binary_str(prefix_ip)
        roa_tree = self.roa_tree6 if ip_version == 6 else self.roa_tree4
        roas_covering = roa_tree.prefix_items(ip_bin_str[:prefix_length])
        if not roas_covering:
            return RPKIStatus.not_found
        for key, value in roas_covering:
            for roa_prefix, roa_asn, roa_max_length in value:
                if roa_asn != 0 and roa_asn == prefix_asn and prefix_length <= roa_max_length:
                    return RPKIStatus.valid
        return RPKIStatus.invalid

    def _build_roa_tree_from_roa_objs(self, roas: List[ROA]):
        """
        Build the tree of all ROAs from ROA objects.
        """
        for roa in roas:
            roa_tree = self.roa_tree6 if roa.prefix.version() == 6 else self.roa_tree4
            key = roa.prefix.strBin()[:roa.prefix.prefixlen()]
            if key in roa_tree:
                roa_tree[key].append((roa.prefix_str, roa.asn, roa.max_length))
            else:
                roa_tree[key] = [(roa.prefix_str, roa.asn, roa.max_length)]

    def _build_roa_tree_from_db(self):
        """
        Build the tree of all ROAs from the DB.
        """
        roas = self.database_handler.execute_query(ROADatabaseObjectQuery())
        for roa in roas:
            first_ip, length = roa['prefix'].split('/')
            ip_version, ip_bin_str = self._ip_to_binary_str(first_ip)
            key = ip_bin_str[:int(length)]
            roa_tree = self.roa_tree6 if ip_version == 6 else self.roa_tree4
            if key in roa_tree:
                roa_tree[key].append((roa['prefix'], roa['asn'], roa['max_length']))
            else:
                roa_tree[key] = [(roa['prefix'], roa['asn'], roa['max_length'])]

    def _ip_to_binary_str(self, ip: str) -> Tuple[int, str]:
        """
        Convert an IP string to a binary string, e.g.
        192.0.2.139 to 11000000000000000000001010001011
        and return the IP version.
        """
        address_family = socket.AF_INET6 if ':' in ip else socket.AF_INET
        ip_bin = socket.inet_pton(address_family, ip)
        ip_bin_str = ''.join([BYTE_BIN[b] for b in ip_bin]) + '0'
        ip_version = 6 if address_family == socket.AF_INET6 else 4
        return ip_version, ip_bin_str


class SingleRouteROAValidator:
    def __init__(self, database_handler: DatabaseHandler):
        self.database_handler = database_handler

    def validate_route(self, route: IP, asn: int, source: str) -> RPKIStatus:
        """
        Validate a route from a particular source.
        """
        if get_setting(f'sources.{source}.rpki_excluded'):
            return RPKIStatus.not_found

        query = ROADatabaseObjectQuery().ip_less_specific_or_exact(route)
        roas_covering = list(self.database_handler.execute_query(query))
        if not roas_covering:
            return RPKIStatus.not_found
        for roa in roas_covering:
            if roa['asn'] != 0 and roa['asn'] == asn and route.prefixlen() <= roa['max_length']:
                return RPKIStatus.valid
        return RPKIStatus.invalid



from typing import Tuple, Union


def parse_as_number(value: Union[str, int], permit_plain=False) -> Tuple[str, int]:
    """Validate and clean an AS number. Returns it in ASxxxx and numeric format."""
    if isinstance(value, str):
        value = value.upper()
        if not permit_plain and not value.startswith('AS'):
            raise ValidationError(f'Invalid AS number {value}: must start with "AS"')

        start_index = 2 if value.startswith('AS') else 0

        if not value[start_index:].isnumeric():
            raise ValidationError(f'Invalid AS number {value}: number part is not numeric')

        value_int = int(value[start_index:])
    else:
        value_int = value

    if not 0 <= value_int <= 4294967295:
        raise ValidationError(f'Invalid AS number {value}: valid range is 0-4294967295')

    return 'AS' + str(value_int), value_int


class ValidationError(ValueError):
    pass



from collections import defaultdict
from typing import Optional, Tuple, List, Dict

from IPy import IP

from irrd.conf import get_setting
from irrd.rpsl.parser import RPSLObject
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery
from .status import ScopeFilterStatus


class ScopeFilterValidator:
    """
    The scope filter validator validates whether prefixes, ASNs or RPSL
    objects fall within the configured scope filter.
    """

    def __init__(self):
        self.load_filters()

    def load_filters(self):
        """
        (Re)load the local cache of the configured filters.
        Also called by __init__
        """
        prefixes = get_setting('scopefilter.prefixes', [])
        self.filtered_prefixes = [IP(prefix) for prefix in prefixes]

        self.filtered_asns = set()
        self.filtered_asn_ranges = set()
        asn_filters = get_setting('scopefilter.asns', [])
        for asn_filter in asn_filters:
            if '-' in str(asn_filter):
                start, end = asn_filter.split('-')
                self.filtered_asn_ranges.add((int(start), int(end)))
            else:
                self.filtered_asns.add(int(asn_filter))

    def validate(self, source: str, prefix: Optional[IP]=None, asn: Optional[int]=None) -> ScopeFilterStatus:
        """
        Validate a prefix and/or ASN, for a particular source.
        Returns a tuple of a ScopeFilterStatus and an explanation string.
        """
        if not prefix and asn is None:
            raise ValueError('Scope Filter validator must be provided asn or prefix')

        if get_setting(f'sources.{source}.scopefilter_excluded'):
            return ScopeFilterStatus.in_scope

        if prefix:
            for filtered_prefix in self.filtered_prefixes:
                if prefix.version() == filtered_prefix.version() and filtered_prefix.overlaps(prefix):
                    return ScopeFilterStatus.out_scope_prefix

        if asn is not None:
            if asn in self.filtered_asns:
                return ScopeFilterStatus.out_scope_as
            for range_start, range_end in self.filtered_asn_ranges:
                if range_start <= asn <= range_end:
                    return ScopeFilterStatus.out_scope_as

        return ScopeFilterStatus.in_scope

    def _validate_rpsl_data(self, source: str, object_class: str, prefix: Optional[IP],
                            asn_first: Optional[int]) -> Tuple[ScopeFilterStatus, str]:
        """
        Validate whether a particular set of RPSL data is in scope.
        Returns a ScopeFilterStatus.
        """
        out_of_scope = [ScopeFilterStatus.out_scope_prefix, ScopeFilterStatus.out_scope_as]
        if object_class not in ['route', 'route6', 'aut-num']:
            return ScopeFilterStatus.in_scope, ''

        if prefix:
            prefix_state = self.validate(source, prefix)
            if prefix_state in out_of_scope:
                return prefix_state, f'prefix {prefix} is out of scope'

        if asn_first is not None:
            asn_state = self.validate(source, asn=asn_first)
            if asn_state in out_of_scope:
                return asn_state, f'ASN {asn_first} is out of scope'

        return ScopeFilterStatus.in_scope, ''

    def validate_rpsl_object(self, rpsl_object: RPSLObject) -> Tuple[ScopeFilterStatus, str]:
        """
        Validate whether an RPSLObject is in scope.
        Returns a tuple of a ScopeFilterStatus and an explanation string.
        """
        return self._validate_rpsl_data(
            rpsl_object.source(),
            rpsl_object.rpsl_object_class,
            rpsl_object.prefix,
            rpsl_object.asn_first,
        )

    def validate_all_rpsl_objects(self, database_handler: DatabaseHandler) -> \
            Tuple[List[Dict[str, str]], List[Dict[str, str]], List[Dict[str, str]]]:
        """
        Apply the scope filter to all relevant objects.

        Retrieves all routes from the DB, and aggregates the validation results.
        Returns a tuple of three sets:
        - one with routes that should be set to status in_scope, but are not now
        - one with routes that should be set to status out_scope_as, but are not now
        - one with routes that should be set to status out_scope_prefix, but are not now
        Each object is recorded as a dict, which has the fields shown
        in "columns" below.

        Objects where their current status in the DB matches the new
        validation result, are not included in the return value.
        """
        columns = ['rpsl_pk', 'ip_first', 'prefix_length', 'asn_first', 'source', 'object_class',
                   'object_text', 'scopefilter_status', 'rpki_status']

        objs_changed: Dict[ScopeFilterStatus, List[Dict[str, str]]] = defaultdict(list)

        q = RPSLDatabaseQuery(column_names=columns, enable_ordering=False)
        q = q.object_classes(['route', 'route6', 'aut-num'])
        results = database_handler.execute_query(q)

        for result in results:
            current_status = result['scopefilter_status']
            result['old_status'] = current_status
            prefix = None
            if result.get('ip_first'):
                prefix = IP(result['ip_first'] + '/' + str(result['prefix_length']))
            new_status, _ = self._validate_rpsl_data(
                result['source'],
                result['object_class'],
                prefix,
                result['asn_first'],
            )
            if new_status != current_status:
                result['scopefilter_status'] = new_status
                objs_changed[new_status].append(result)
        return (objs_changed[ScopeFilterStatus.in_scope],
                objs_changed[ScopeFilterStatus.out_scope_as],
                objs_changed[ScopeFilterStatus.out_scope_prefix])



import enum

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.declarative import declarative_base, declared_attr

from irrd.rpki.status import RPKIStatus
from irrd.rpsl.rpsl_objects import lookup_field_names
from irrd.scopefilter.status import ScopeFilterStatus


class DatabaseOperation(enum.Enum):
    add_or_update = 'ADD'
    delete = 'DEL'


class JournalEntryOrigin(enum.Enum):
    # Legacy journal entries for which the origin is unknown, can be auth_change or mirror
    unknown = 'UNKNOWN'
    # Journal entry received from a mirror by NRTM or importing from a file
    mirror = 'MIRROR'
    # Journal entry generated from synthesized NRTM
    synthetic_nrtm = 'SYNTHETIC_NRTM'
    # Journal entry generated from pseudo IRR
    pseudo_irr = 'PSEUDO_IRR'
    # Journal entry caused by a user-submitted change in an authoritative database
    auth_change = 'AUTH_CHANGE'
    # Journal entry caused by a change in in the RPKI status of an object in an authoritative db
    rpki_status = 'RPKI_STATUS'
    # Journal entry caused by a change in the scope filter status
    scope_filter = 'SCOPE_FILTER'


Base = declarative_base()


class RPSLDatabaseObject(Base):  # type: ignore
    """
    SQLAlchemy ORM object for RPSL database objects.

    Note that SQLAlchemy does not require you to use the ORM for ORM
    objects - as that can be slower with large queries.
    """
    __tablename__ = 'rpsl_objects'

    # Requires extension pgcrypto
    # in alembic: op.execute('create EXTENSION if not EXISTS 'pgcrypto';')
    pk = sa.Column(pg.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'), primary_key=True)
    rpsl_pk = sa.Column(sa.String, index=True, nullable=False)
    source = sa.Column(sa.String, index=True, nullable=False)

    object_class = sa.Column(sa.String, nullable=False, index=True)
    parsed_data = sa.Column(pg.JSONB, nullable=False)
    object_text = sa.Column(sa.Text, nullable=False)

    ip_version = sa.Column(sa.Integer, index=True)
    ip_first = sa.Column(pg.INET, index=True)
    ip_last = sa.Column(pg.INET, index=True)
    ip_size = sa.Column(sa.DECIMAL(scale=0))
    # Only filled for route/route6
    prefix_length = sa.Column(sa.Integer, nullable=True)

    asn_first = sa.Column(sa.BigInteger, index=True)
    asn_last = sa.Column(sa.BigInteger, index=True)

    rpki_status = sa.Column(sa.Enum(RPKIStatus), nullable=False, index=True, server_default=RPKIStatus.not_found.name)
    scopefilter_status = sa.Column(sa.Enum(ScopeFilterStatus), nullable=False, index=True, server_default=ScopeFilterStatus.in_scope.name)

    created = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False)
    updated = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False)

    @declared_attr
    def __table_args__(cls):  # noqa
        args = [
            sa.UniqueConstraint('rpsl_pk', 'source', name='rpsl_objects_rpsl_pk_source_unique'),
            sa.Index('ix_rpsl_objects_ip_first_ip_last', 'ip_first', 'ip_last', ),
            sa.Index('ix_rpsl_objects_ip_last_ip_first', 'ip_last', 'ip_first'),
            sa.Index('ix_rpsl_objects_asn_first_asn_last', 'asn_first', 'asn_last'),
        ]
        for name in lookup_field_names():
            index_name = 'ix_rpsl_objects_parsed_data_' + name.replace('-', '_')
            index_on = sa.text(f"(parsed_data->'{name}')")
            args.append(sa.Index(index_name, index_on, postgresql_using='gin'))
        return tuple(args)

    def __repr__(self):
        return f'<{self.rpsl_pk}/{self.source}/{self.pk}>'


class RPSLDatabaseJournal(Base):  # type: ignore
    """
    SQLAlchemy ORM object for change history of RPSL database objects.
    """
    __tablename__ = 'rpsl_database_journal'

    # Requires extension pgcrypto
    # in alembic: op.execute('create EXTENSION if not EXISTS 'pgcrypto';')
    pk = sa.Column(pg.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'), primary_key=True)
    rpsl_pk = sa.Column(sa.String, index=True, nullable=False)
    source = sa.Column(sa.String, index=True, nullable=False)
    origin = sa.Column(sa.Enum(JournalEntryOrigin), nullable=False, index=True, server_default=JournalEntryOrigin.unknown.name)

    serial_nrtm = sa.Column(sa.Integer, index=True, nullable=False)
    operation = sa.Column(sa.Enum(DatabaseOperation), nullable=False)

    object_class = sa.Column(sa.String, nullable=False, index=True)
    object_text = sa.Column(sa.Text, nullable=False)

    # These objects are not mutable, so creation time is sufficient.
    timestamp = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False)

    @declared_attr
    def __table_args__(cls):  # noqa
        return (
            sa.UniqueConstraint('serial_nrtm', 'source', name='rpsl_objects_history_serial_nrtm_source_unique'),
        )

    def __repr__(self):
        return f'<{self.source}/{self.serial}/{self.operation}/{self.rpsl_pk}>'


class RPSLDatabaseStatus(Base):  # type: ignore
    """
    SQLAlchemy ORM object for the status of authoritative and mirrored DBs.

    Note that this database is for keeping status, and is not the source
    of configuration parameters.
    """
    __tablename__ = 'database_status'

    pk = sa.Column(pg.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'), primary_key=True)
    source = sa.Column(sa.String, index=True, nullable=False, unique=True)

    # The oldest and newest serials seen, for any reason since the last import
    serial_oldest_seen = sa.Column(sa.Integer)
    serial_newest_seen = sa.Column(sa.Integer)
    # The oldest and newest serials in the current local journal
    serial_oldest_journal = sa.Column(sa.Integer)
    serial_newest_journal = sa.Column(sa.Integer)
    # The serial at which this IRRd last exported
    serial_last_export = sa.Column(sa.Integer)
    # The last serial seen from an NRTM mirror, i.e. resume NRTM query from this serial
    serial_newest_mirror = sa.Column(sa.Integer)

    force_reload = sa.Column(sa.Boolean(), default=False, nullable=False)
    synchronised_serials = sa.Column(sa.Boolean(), default=True, nullable=False)

    last_error = sa.Column(sa.Text)
    last_error_timestamp = sa.Column(sa.DateTime(timezone=True))

    created = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False)
    updated = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), onupdate=sa.func.now(), nullable=False)

    def __repr__(self):
        return self.source


class ROADatabaseObject(Base):  # type: ignore
    """
    SQLAlchemy ORM object for ROA objects.
    """
    __tablename__ = 'roa_object'

    pk = sa.Column(pg.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'), primary_key=True)

    prefix = sa.Column(pg.CIDR, nullable=False)
    asn = sa.Column(sa.BigInteger, nullable=False)
    max_length = sa.Column(sa.Integer, nullable=False)
    trust_anchor = sa.Column(sa.String)
    ip_version = sa.Column(sa.Integer, nullable=False, index=True)

    created = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False)

    @declared_attr
    def __table_args__(cls):  # noqa
        args = [
            sa.UniqueConstraint('prefix', 'asn', 'max_length', 'trust_anchor', name='roa_object_prefix_asn_maxlength_unique'),
            sa.Index('ix_roa_objects_prefix_gist', sa.text('prefix inet_ops'), postgresql_using='gist')
        ]
        return tuple(args)

    def __repr__(self):
        return f'<{self.prefix}/{self.asn}>'


# Before you update this, please check the storage documentation for changing lookup fields.
expected_lookup_field_names = {
    'admin-c', 'tech-c', 'zone-c', 'member-of', 'mnt-by', 'role', 'members', 'person',
    'mp-members', 'origin', 'mbrs-by-ref',
}
if sorted(lookup_field_names()) != sorted(expected_lookup_field_names):  # pragma: no cover
    raise RuntimeError(f'Field names of lookup fields do not match expected set. Indexes may be missing. '
                       f'Expected: {expected_lookup_field_names}, actual: {lookup_field_names()}')



# flake8: noqa: E402
import os
import sys

from alembic import context
from pathlib import Path
from sqlalchemy import create_engine

sys.path.append(str(Path(__file__).resolve().parents[3]))

from irrd.conf import get_setting, config_init, is_config_initialised
from irrd.storage.models import Base

target_metadata = Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    if not is_config_initialised():
        config_init(os.environ['IRRD_CONFIG_FILE'])
    url = get_setting('database_url')
    context.configure(
        url=url, target_metadata=target_metadata, literal_binds=True,
        transaction_per_migration=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    if not is_config_initialised():
        config_init(os.environ['IRRD_CONFIG_FILE'])
    engine = create_engine(get_setting('database_url'))

    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import logging
import re
from typing import List, Set, Optional

from irrd.conf import get_setting
from irrd.rpki.validators import BulkRouteROAValidator
from irrd.rpsl.parser import UnknownRPSLObjectClassException, RPSLObject
from irrd.rpsl.rpsl_objects import rpsl_object_from_text, RPSLKeyCert
from irrd.scopefilter.validators import ScopeFilterValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.models import DatabaseOperation, JournalEntryOrigin
from irrd.utils.text import split_paragraphs_rpsl, remove_last_modified
from .nrtm_operation import NRTMOperation
from ..storage.queries import RPSLDatabaseQuery

logger = logging.getLogger(__name__)
nrtm_start_line_re = re.compile(r'^% *START *Version: *(?P<version>\d+) +(?P<source>[\w-]+) +(?P<first_serial>\d+)-(?P<last_serial>\d+)( FILTERED)?\n$', flags=re.MULTILINE)


class RPSLImportError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class MirrorParser:
    def __init__(self):
        object_class_filter = get_setting(f'sources.{self.source}.object_class_filter')
        if object_class_filter:
            if isinstance(object_class_filter, str):
                object_class_filter = [object_class_filter]
            self.object_class_filter = [c.strip().lower() for c in object_class_filter]
        else:
            self.object_class_filter = None

        self.strict_validation_key_cert = get_setting(f'sources.{self.source}.strict_import_keycert_objects', False)


class MirrorFileImportParserBase(MirrorParser):
    """
    This parser handles imports of files for mirror databases.
    Note that this parser can be called multiple times for a single
    full import, as some databases use split files.

    If direct_error_return is set, run_import() immediately returns
    upon an encountering an error message. It will return an error
    string.
    """
    obj_parsed = 0  # Total objects found
    obj_errors = 0  # Objects with errors
    obj_ignored_class = 0  # Objects ignored due to object_class_filter setting
    obj_unknown = 0  # Objects with unknown classes
    unknown_object_classes: Set[str] = set()  # Set of encountered unknown classes

    def __init__(self, source: str, filename: str,
                 database_handler: DatabaseHandler, direct_error_return: bool=False,
                 roa_validator: Optional[BulkRouteROAValidator] = None) -> None:
        self.source = source
        self.filename = filename
        self.database_handler = database_handler
        self.direct_error_return = direct_error_return
        self.roa_validator = roa_validator
        self.obj_parsed = 0  # Total objects found
        self.obj_errors = 0  # Objects with errors
        self.obj_ignored_class = 0  # Objects ignored due to object_class_filter setting
        self.obj_unknown = 0  # Objects with unknown classes
        self.unknown_object_classes: Set[str] = set()  # Set of encountered unknown classes
        self.scopefilter_validator = ScopeFilterValidator()
        super().__init__()

    def _parse_object(self, rpsl_text: str) -> Optional[RPSLObject]:
        """
        Parse and validate a single object and return it.
        If there is a parsing error, unknown object class, invalid source:
        - if direct_error_return is set, raises an RPSLImportError
        - otherwise, returns None
        """
        try:
            self.obj_parsed += 1
            # If an object turns out to be a key-cert, and strict_import_keycert_objects
            # is set, parse it again with strict validation to load it in the GPG keychain.
            obj = rpsl_object_from_text(rpsl_text.strip(), strict_validation=False)
            if self.strict_validation_key_cert and obj.__class__ == RPSLKeyCert:
                obj = rpsl_object_from_text(rpsl_text.strip(), strict_validation=True)

            if obj.messages.errors():
                log_msg = f'Parsing errors: {obj.messages.errors()}, original object text follows:\n{rpsl_text}'
                if self.direct_error_return:
                    raise RPSLImportError(log_msg)
                self.database_handler.record_mirror_error(self.source, log_msg)
                logger.critical(f'Parsing errors occurred while importing from file for {self.source}. '
                                f'This object is ignored, causing potential data inconsistencies. A new operation for '
                                f'this update, without errors, will still be processed and cause the inconsistency to '
                                f'be resolved. Parser error messages: {obj.messages.errors()}; '
                                f'original object text follows:\n{rpsl_text}')
                self.obj_errors += 1
                return None

            if obj.source() != self.source:
                msg = f'Invalid source {obj.source()} for object {obj.pk()}, expected {self.source}'
                if self.direct_error_return:
                    raise RPSLImportError(msg)
                logger.critical(msg + '. This object is ignored, causing potential data inconsistencies.')
                self.database_handler.record_mirror_error(self.source, msg)
                self.obj_errors += 1
                return None

            if self.object_class_filter and obj.rpsl_object_class.lower() not in self.object_class_filter:
                self.obj_ignored_class += 1
                return None

            if self.roa_validator and obj.rpki_relevant and obj.prefix_length and obj.asn_first:
                obj.rpki_status = self.roa_validator.validate_route(
                    str(obj.ip_first), obj.prefix_length, obj.asn_first, obj.source()
                )

            obj.scopefilter_status, _ = self.scopefilter_validator.validate_rpsl_object(obj)

            return obj

        except UnknownRPSLObjectClassException as e:
            # Ignore legacy IRRd artifacts
            # https://github.com/irrdnet/irrd4/issues/232
            if e.rpsl_object_class.startswith('*xx'):
                self.obj_parsed -= 1  # This object does not exist to us
                return None
            if self.direct_error_return:
                raise RPSLImportError(f'Unknown object class: {e.rpsl_object_class}')
            self.obj_unknown += 1
            self.unknown_object_classes.add(e.rpsl_object_class)
        return None


class MirrorFileImportParser(MirrorFileImportParserBase):
    """
    This parser handles imports of files for mirror databases.
    Note that this parser can be called multiple times for a single
    full import, as some databases use split files.

    If direct_error_return is set, run_import() immediately returns
    upon an encountering an error message. It will return an error
    string.
    """
    def __init__(self, serial: Optional[int]=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serial = serial
        logger.debug(f'Starting file import of {self.source} from {self.filename}')

    def run_import(self) -> Optional[str]:
        """
        Run the actual import. If direct_error_return is set, returns an error
        string on encountering the first error. Otherwise, returns None.
        """
        f = open(self.filename, encoding='utf-8', errors='backslashreplace')
        for paragraph in split_paragraphs_rpsl(f):
            try:
                rpsl_obj = self._parse_object(paragraph)
            except RPSLImportError as e:
                if self.direct_error_return:
                    return e.message
            else:
                if rpsl_obj:
                    self.database_handler.upsert_rpsl_object(
                        rpsl_obj, origin=JournalEntryOrigin.mirror)

        self.log_report()
        f.close()
        if self.serial:
            self.database_handler.record_serial_seen(self.source, self.serial)

        return None

    def log_report(self) -> None:
        obj_successful = self.obj_parsed - self.obj_unknown - self.obj_errors - self.obj_ignored_class
        logger.info(f'File import for {self.source}: {self.obj_parsed} objects read, '
                    f'{obj_successful} objects inserted, '
                    f'ignored {self.obj_errors} due to errors, '
                    f'ignored {self.obj_ignored_class} due to object_class_filter, '
                    f'source {self.filename}')
        if self.obj_unknown:
            unknown_formatted = ', '.join(self.unknown_object_classes)
            logger.warning(f'Ignored {self.obj_unknown} objects found in file import for {self.source} due to unknown '
                           f'object classes: {unknown_formatted}')


class MirrorUpdateFileImportParser(MirrorFileImportParserBase):
    """
    This parser handles files for mirror databases, and processes them
    as the new state of that mirror, including journal entries if needed.
    Note that this parser *can not* be called multiple times for a single
    full import, i.e. it is not compatible with split files - when given
    a particular file, that is treated as the full current state of that
    particular source.

    If direct_error_return is set, run_import() immediately returns
    upon an encountering an error message. It will return an error
    string.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.debug(f'Starting update import for {self.source} from {self.filename}')
        self.obj_new = 0  # New objects
        self.obj_modified = 0  # Modified objects
        self.obj_retained = 0  # Retained and possibly modified objects
        self.obj_deleted = 0  # Deleted objects

    def run_import(self) -> Optional[str]:
        """
        Run the actual import. If direct_error_return is set, returns an error
        string on encountering the first error. Otherwise, returns None.
        """
        objs_from_file = []
        f = open(self.filename, encoding='utf-8', errors='backslashreplace')
        for paragraph in split_paragraphs_rpsl(f):
            try:
                rpsl_obj = self._parse_object(paragraph)
            except RPSLImportError as e:
                if self.direct_error_return:
                    return e.message
            else:
                if rpsl_obj:
                    objs_from_file.append(rpsl_obj)
        f.close()

        query = RPSLDatabaseQuery(ordered_by_sources=False, enable_ordering=False,
                                  column_names=['rpsl_pk']).sources([self.source])
        current_pks = {row['rpsl_pk'] for row in self.database_handler.execute_query(query)}

        file_objs_by_pk = {obj.pk(): obj for obj in objs_from_file}
        file_pks = set(file_objs_by_pk.keys())
        new_pks = file_pks - current_pks
        deleted_pks = current_pks - file_pks
        retained_pks = file_pks.intersection(current_pks)

        self.obj_new = len(new_pks)
        self.obj_deleted = len(deleted_pks)
        self.obj_retained = len(retained_pks)

        for rpsl_pk, file_obj in filter(lambda i: i[0] in new_pks, file_objs_by_pk.items()):
            self.database_handler.upsert_rpsl_object(file_obj, JournalEntryOrigin.synthetic_nrtm)

        for rpsl_pk in deleted_pks:
            self.database_handler.delete_rpsl_object(rpsl_pk=rpsl_pk, source=self.source,
                                                     origin=JournalEntryOrigin.synthetic_nrtm)

        # This query does not filter on retained_pks. The expectation is that most
        # objects are retained, and therefore it is much faster to query the entire source.
        query = RPSLDatabaseQuery(ordered_by_sources=False, enable_ordering=False,
                                  column_names=['rpsl_pk', 'object_text'])
        query = query.sources([self.source])
        for row in self.database_handler.execute_query(query):
            try:
                file_obj = file_objs_by_pk[row['rpsl_pk']]
            except KeyError:
                continue
            if file_obj.render_rpsl_text() != remove_last_modified(row['object_text']):
                self.database_handler.upsert_rpsl_object(file_obj, JournalEntryOrigin.synthetic_nrtm)
                self.obj_modified += 1

        self.log_report()
        return None

    def log_report(self) -> None:
        obj_successful = self.obj_parsed - self.obj_unknown - self.obj_errors - self.obj_ignored_class
        logger.info(f'File update for {self.source}: {self.obj_parsed} objects read, '
                    f'{obj_successful} objects processed, '
                    f'{self.obj_new} objects newly inserted, '
                    f'{self.obj_deleted} objects newly deleted, '
                    f'{self.obj_retained} objects retained, of which '
                    f'{self.obj_modified} modified, '
                    f'ignored {self.obj_errors} due to errors, '
                    f'ignored {self.obj_ignored_class} due to object_class_filter, '
                    f'source {self.filename}')
        if self.obj_unknown:
            unknown_formatted = ', '.join(self.unknown_object_classes)
            logger.warning(f'Ignored {self.obj_unknown} objects found in file import for {self.source} due to unknown '
                           f'object classes: {unknown_formatted}')


class NRTMStreamParser(MirrorParser):
    """
    The NRTM parser takes the data of an NRTM string, and splits it
    into individual operations, matched with their serial and
    whether they are an ADD/DEL operation.

    Creating an instance will fill the attributes:
    - first_serial: the first serial found in the data
    - last_serial: the last serial found
    - nrtm_source: the RPSL source recorded in the START header (must be equal to expected source)
    - operations: a list of NRTMOperation objects

    Raises a ValueError for invalid NRTM data.
    """
    first_serial = -1
    last_serial = -1
    nrtm_source = None
    _current_op_serial = -1

    def __init__(self, source: str, nrtm_data: str, database_handler: DatabaseHandler) -> None:
        self.source = source
        self.database_handler = database_handler
        self.rpki_aware = bool(get_setting('rpki.roa_source'))
        super().__init__()
        self.operations: List[NRTMOperation] = []
        self._split_stream(nrtm_data)

    def _split_stream(self, data: str) -> None:
        """Split a stream into individual operations."""
        paragraphs = split_paragraphs_rpsl(data, strip_comments=False)
        last_comment_seen = ''

        for paragraph in paragraphs:
            if self._handle_possible_start_line(paragraph):
                continue
            elif paragraph.startswith('%') or paragraph.startswith('#'):
                last_comment_seen = paragraph
            elif paragraph.startswith('ADD') or paragraph.startswith('DEL'):
                self._handle_operation(paragraph, paragraphs)

        if self.nrtm_source and last_comment_seen.upper().strip() != f'%END {self.source}':
            msg = f'NRTM stream error for {self.source}: last paragraph expected to be ' \
                  f'"%END {self.source}", but is actually {last_comment_seen.upper().strip()}'
            logger.error(msg)
            self.database_handler.record_mirror_error(self.source, msg)
            raise ValueError(msg)

        if self._current_op_serial > self.last_serial and self.version != '3':
            msg = f'NRTM stream error for {self.source}: expected operations up to and including serial ' \
                  f'{self.last_serial}, last operation was {self._current_op_serial}'
            logger.error(msg)
            self.database_handler.record_mirror_error(self.source, msg)
            raise ValueError(msg)

        if self.last_serial > 0:
            self.database_handler.record_serial_newest_mirror(self.source, self.last_serial)

    def _handle_possible_start_line(self, line: str) -> bool:
        """Check whether a line is an NRTM START line, and if so, handle it."""
        start_line_match = nrtm_start_line_re.match(line)
        if not start_line_match:
            return False

        if self.nrtm_source:  # nrtm_source can only be defined if this is a second START line
            msg = f'Encountered second START line in NRTM stream, first was {self.source} ' \
                  f'{self.first_serial}-{self.last_serial}, new line is: {line}'
            self.database_handler.record_mirror_error(self.source, msg)
            logger.error(msg)
            raise ValueError(msg)

        self.version = start_line_match.group('version')
        self.nrtm_source = start_line_match.group('source').upper()
        self.first_serial = int(start_line_match.group('first_serial'))
        self.last_serial = int(start_line_match.group('last_serial'))

        if self.source != self.nrtm_source:
            msg = f'Invalid NRTM source in START line: expected {self.source} but found ' \
                  f'{self.nrtm_source} in line: {line}'
            self.database_handler.record_mirror_error(self.source, msg)
            logger.error(msg)
            raise ValueError(msg)

        if self.version not in ['1', '3']:
            msg = f'Invalid NRTM version {self.version} in START line: {line}'
            self.database_handler.record_mirror_error(self.source, msg)
            logger.error(msg)
            raise ValueError(msg)

        logger.debug(f'Found valid start line for {self.source}, range {self.first_serial}-{self.last_serial}')

        return True

    def _handle_operation(self, current_paragraph: str, paragraphs) -> None:
        """Handle a single ADD/DEL operation."""
        if not self.nrtm_source:
            msg = f'Encountered operation before valid NRTM START line, paragraph encountered: {current_paragraph}'
            self.database_handler.record_mirror_error(self.source, msg)
            logger.error(msg)
            raise ValueError(msg)

        if self._current_op_serial == -1:
            self._current_op_serial = self.first_serial
        else:
            self._current_op_serial += 1

        if ' ' in current_paragraph:
            operation_str, line_serial_str = current_paragraph.split(' ')
            line_serial = int(line_serial_str)
            # Gaps are allowed, but the line serial can never be lower, as that
            # means operations are served in the wrong order.
            if line_serial < self._current_op_serial:
                msg = f'Invalid NRTM serial for {self.source}: ADD/DEL has serial {line_serial}, ' \
                      f'expected at least {self._current_op_serial}'
                logger.error(msg)
                self.database_handler.record_mirror_error(self.source, msg)
                raise ValueError(msg)
            self._current_op_serial = line_serial
        else:
            operation_str = current_paragraph.strip()

        operation = DatabaseOperation(operation_str)
        object_text = next(paragraphs)
        nrtm_operation = NRTMOperation(self.source, operation, self._current_op_serial,
                                       object_text, self.strict_validation_key_cert, self.rpki_aware,
                                       self.object_class_filter)
        self.operations.append(nrtm_operation)



import logging
from collections import defaultdict
from datetime import datetime, timezone
from functools import lru_cache
from io import StringIO
from typing import List, Set, Dict, Any, Tuple, Iterator, Union, Optional

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg

from irrd.conf import get_setting
from irrd.rpki.status import RPKIStatus
from irrd.rpsl.parser import RPSLObject
from irrd.rpsl.rpsl_objects import OBJECT_CLASS_MAPPING
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.vendor import postgres_copy
from . import get_engine
from .models import RPSLDatabaseObject, RPSLDatabaseJournal, DatabaseOperation, RPSLDatabaseStatus, \
    ROADatabaseObject, JournalEntryOrigin
from .preload import Preloader
from .queries import (BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery,
                      RPSLDatabaseObjectStatisticsQuery, ROADatabaseObjectQuery)

logger = logging.getLogger(__name__)
MAX_RECORDS_BUFFER_BEFORE_INSERT = 15000


class DatabaseHandler:
    """
    Interface for other parts of IRRD to talk to the database.

    Note that no writes to the database are final until commit()
    has been called - and rollback() can be called at any time
    to all submitted changes.
    """
    journaling_enabled: bool
    _transaction: sa.engine.base.Transaction
    _rpsl_pk_source_seen: Set[str]
    # The RPSL upsert buffer is a list of tuples. Each tuple first has a dict
    # with all database column names and their values, and then the origin of the change
    # and the serial of the change at the NRTM source, if any.
    _rpsl_upsert_buffer: List[Tuple[dict, JournalEntryOrigin, Optional[int]]]
    # The ROA insert buffer is a list of dicts with columm names and their values.
    _roa_insert_buffer: List[Dict[str, Union[str, int]]]

    def __init__(self, readonly=False):
        """
        Create a new database handler.

        If readonly is True, this instance will expect read queries only.
        No transaction will be started, all queries will use autocommit.
        Readonly is always true if database_readonly is set in the config.
        """
        if get_setting('database_readonly'):
            self.readonly = True
        else:
            self.readonly = readonly
        self.journaling_enabled = not readonly
        self._connection = get_engine().connect()
        if self.readonly:
            self._connection.execution_options(isolation_level="AUTOCOMMIT")
        else:
            self._start_transaction()
            self.preloader = Preloader(enable_queries=False)

    def refresh_connection(self) -> None:
        """
        Refresh the database connection.
        Needed when trying to re-use this instance when the
        SQL server has restarted.
        """
        try:
            if not self.readonly:
                self.rollback(start_transaction=False)
        except Exception:  # pragma: no cover
            pass
        try:
            self.close()
        except Exception:  # pragma: no cover
            pass
        self._connection = get_engine().connect()
        if self.readonly:
            self._connection.execution_options(isolation_level="AUTOCOMMIT")
        else:
            self._start_transaction()

    def _start_transaction(self) -> None:
        """Start a fresh transaction."""
        self._transaction = self._connection.begin()
        self._rpsl_pk_source_seen: Set[str] = set()
        self._rpsl_upsert_buffer = []
        self._roa_insert_buffer = []
        self._object_classes_modified: Set[str] = set()
        self._rpsl_guaranteed_no_existing = True
        self.status_tracker = DatabaseStatusTracker(self, journaling_enabled=self.journaling_enabled)

    def disable_journaling(self):
        self.journaling_enabled = False
        self.status_tracker.journaling_enabled = False

    def enable_journaling(self):
        self.journaling_enabled = True
        self.status_tracker.journaling_enabled = True

    def commit(self) -> None:
        """
        Commit any pending changes to the database and start a fresh transaction.
        """
        self._check_write_permitted()
        self._flush_rpsl_object_writing_buffer()
        self._flush_roa_writing_buffer()
        self.status_tracker.finalise_transaction()
        try:
            self._transaction.commit()
            if self._object_classes_modified:
                self.preloader.signal_reload(self._object_classes_modified)
            self._start_transaction()
        except Exception as exc:  # pragma: no cover
            self._transaction.rollback()
            logger.error('Exception occurred while committing changes, rolling back', exc_info=exc)
            raise

    def rollback(self, start_transaction=True) -> None:
        """Roll back the current transaction, discarding all submitted changes."""
        self._rpsl_upsert_buffer = []
        self._rpsl_pk_source_seen = set()
        self.status_tracker.reset()
        self._transaction.rollback()
        if start_transaction:
            self._start_transaction()

    def execute_query(self,
                      query: Union[BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery, RPSLDatabaseObjectStatisticsQuery, ROADatabaseObjectQuery],
                      flush_rpsl_buffer=True,
                      ) -> Iterator[Dict[str, Any]]:
        """
        Execute an RPSLDatabaseQuery within the current transaction.
        If flush_rpsl_buffer is set, the RPSL object buffer is flushed first.
        """
        # To be able to query objects that were just created, flush the buffer.
        if not self.readonly and flush_rpsl_buffer:
            self._flush_rpsl_object_writing_buffer()
        statement = query.finalise_statement()
        result = self._connection.execute(statement)
        for row in result.fetchall():
            yield dict(row)
        result.close()

    def execute_statement(self, statement):
        """Execute a raw SQLAlchemy statement, without flushing the upsert buffer."""
        return self._connection.execute(statement)

    def upsert_rpsl_object(self,
                           rpsl_object: RPSLObject,
                           origin: JournalEntryOrigin,
                           rpsl_guaranteed_no_existing=False,
                           source_serial: Optional[int]=None) -> None:
        """
        Schedule an RPSLObject for insertion/updating.

        This method will insert the object, or overwrite an existing object
        if it has the same RPSL primary key and source. No other checks are
        applied before overwriting.

        Writes may not be issued to the database immediately for performance
        reasons, but commit() will ensure all writes are flushed to the DB first.

        The origin indicates the origin of this change, see JournalEntryOrigin
        for the various options. The source_serial is the serial that an NRTM
        source assigned to this change, if any.

        If rpsl_guaranteed_no_existing is set to True, the caller guarantees that this
        PK is unique in the database. This essentially only applies to inserting
        RPKI psuedo-IRR objects.
        """
        self._check_write_permitted()
        if not rpsl_guaranteed_no_existing:
            self._rpsl_guaranteed_no_existing = False
        ip_first = str(rpsl_object.ip_first) if rpsl_object.ip_first else None
        ip_last = str(rpsl_object.ip_last) if rpsl_object.ip_last else None

        ip_size = None
        if rpsl_object.ip_first and rpsl_object.ip_last:
            ip_size = rpsl_object.ip_last.int() - rpsl_object.ip_first.int() + 1

        # In some cases, multiple updates may be submitted for the same object.
        # PostgreSQL will not allow rows proposed for insertion to have duplicate
        # constrained values - so if a second object appears with a pk/source
        # seen before, the buffer must be flushed right away, or the two updates
        # will conflict.
        source = rpsl_object.parsed_data['source']

        rpsl_pk_source = rpsl_object.pk() + '-' + source
        if rpsl_pk_source in self._rpsl_pk_source_seen:
            self._flush_rpsl_object_writing_buffer()

        update_time = datetime.now(timezone.utc)
        object_dict = {
            'rpsl_pk': rpsl_object.pk(),
            'source': source,
            'object_class': rpsl_object.rpsl_object_class,
            'parsed_data': rpsl_object.parsed_data,
            'object_text': rpsl_object.render_rpsl_text(last_modified=update_time),
            'ip_version': rpsl_object.ip_version(),
            'ip_first': ip_first,
            'ip_last': ip_last,
            'ip_size': ip_size,
            'prefix_length': rpsl_object.prefix_length,
            'asn_first': rpsl_object.asn_first,
            'asn_last': rpsl_object.asn_last,
            'rpki_status': rpsl_object.rpki_status,
            'scopefilter_status': rpsl_object.scopefilter_status,
            'updated': update_time,
        }

        self._rpsl_upsert_buffer.append((object_dict, origin, source_serial))

        self._rpsl_pk_source_seen.add(rpsl_pk_source)
        self._object_classes_modified.add(rpsl_object.rpsl_object_class)

        if len(self._rpsl_upsert_buffer) > MAX_RECORDS_BUFFER_BEFORE_INSERT:
            self._flush_rpsl_object_writing_buffer()

    def insert_roa_object(self, ip_version: int, prefix_str: str, asn: int, max_length: int, trust_anchor: str) -> None:
        """
        Schedule a ROA for insertion.

        Writes for ROAs are only issued when commit() is called,
        as they use a single COPY command. This is possible because
        ROA objects are never updated, only inserted.
        """
        self._check_write_permitted()
        self._roa_insert_buffer.append({
            'ip_version': ip_version,
            'prefix': prefix_str,
            'asn': asn,
            'max_length': max_length,
            'trust_anchor': trust_anchor,
        })

    def update_rpki_status(self, rpsl_objs_now_valid: List[Dict[str, str]]=[],
                           rpsl_objs_now_invalid: List[Dict[str, str]]=[],
                           rpsl_objs_now_not_found: List[Dict[str, str]]=[]) -> None:
        """
        Update the RPKI status for the given RPSL PKs.
        Only PKs whose status have changed should be included.

        Objects that moved to or from invalid generate a journal
        entry, so that mirrors follow the (in)visibility depending
        on RPKI status.
        """
        self._check_write_permitted()
        table = RPSLDatabaseObject.__table__
        if rpsl_objs_now_valid:
            pks = {o['pk'] for o in rpsl_objs_now_valid}
            stmt = table.update().where(table.c.pk.in_(pks)).values(rpki_status=RPKIStatus.valid)
            self.execute_statement(stmt)
        if rpsl_objs_now_invalid:
            pks = {o['pk'] for o in rpsl_objs_now_invalid}
            stmt = table.update().where(table.c.pk.in_(pks)).values(rpki_status=RPKIStatus.invalid)
            self.execute_statement(stmt)
        if rpsl_objs_now_not_found:
            pks = {o['pk'] for o in rpsl_objs_now_not_found}
            stmt = table.update().where(table.c.pk.in_(pks)).values(rpki_status=RPKIStatus.not_found)
            self.execute_statement(stmt)

        for rpsl_obj in rpsl_objs_now_valid + rpsl_objs_now_not_found:
            if all([
                rpsl_obj['old_status'] == RPKIStatus.invalid,
                rpsl_obj['scopefilter_status'] == ScopeFilterStatus.in_scope
            ]):
                self.status_tracker.record_operation(
                    operation=DatabaseOperation.add_or_update,
                    rpsl_pk=rpsl_obj['rpsl_pk'],
                    source=rpsl_obj['source'],
                    object_class=rpsl_obj['object_class'],
                    object_text=rpsl_obj['object_text'],
                    origin=JournalEntryOrigin.rpki_status,
                    source_serial=None,
                )
        for rpsl_obj in rpsl_objs_now_invalid:
            self.status_tracker.record_operation(
                operation=DatabaseOperation.delete,
                rpsl_pk=rpsl_obj['rpsl_pk'],
                source=rpsl_obj['source'],
                object_class=rpsl_obj['object_class'],
                object_text=rpsl_obj['object_text'],
                origin=JournalEntryOrigin.rpki_status,
                source_serial=None,
            )
        if rpsl_objs_now_valid or rpsl_objs_now_invalid or rpsl_objs_now_not_found:
            self._object_classes_modified.add('route')

    def update_scopefilter_status(self, rpsl_objs_now_in_scope: List[Dict[str, str]]=[],
                                  rpsl_objs_now_out_scope_as: List[Dict[str, str]]=[],
                                  rpsl_objs_now_out_scope_prefix: List[Dict[str, str]]=[]) -> None:
        """
        Update the scopefilter status for the given RPSL PKs.
        Only PKs whose status have changed should be included.

        Objects that moved to or from in_scope generate a journal
        entry, so that mirrors follow the (in)visibility depending
        on scopefilter status.
        """
        self._check_write_permitted()
        table = RPSLDatabaseObject.__table__
        if rpsl_objs_now_in_scope:
            pks = {o['rpsl_pk'] for o in rpsl_objs_now_in_scope}
            stmt = table.update().where(table.c.rpsl_pk.in_(pks)).values(scopefilter_status=ScopeFilterStatus.in_scope)
            self.execute_statement(stmt)
        if rpsl_objs_now_out_scope_as:
            pks = {o['rpsl_pk'] for o in rpsl_objs_now_out_scope_as}
            stmt = table.update().where(table.c.rpsl_pk.in_(pks)).values(scopefilter_status=ScopeFilterStatus.out_scope_as)
            self.execute_statement(stmt)
        if rpsl_objs_now_out_scope_prefix:
            pks = {o['rpsl_pk'] for o in rpsl_objs_now_out_scope_prefix}
            stmt = table.update().where(table.c.rpsl_pk.in_(pks)).values(scopefilter_status=ScopeFilterStatus.out_scope_prefix)
            self.execute_statement(stmt)

        for rpsl_obj in rpsl_objs_now_in_scope:
            if rpsl_obj['rpki_status'] != RPKIStatus.invalid:
                self.status_tracker.record_operation(
                    operation=DatabaseOperation.add_or_update,
                    rpsl_pk=rpsl_obj['rpsl_pk'],
                    source=rpsl_obj['source'],
                    object_class=rpsl_obj['object_class'],
                    object_text=rpsl_obj['object_text'],
                    origin=JournalEntryOrigin.scope_filter,
                    source_serial=None,
                )
                self._object_classes_modified.add(rpsl_obj['object_class'])

        for rpsl_obj in rpsl_objs_now_out_scope_as + rpsl_objs_now_out_scope_prefix:
            if rpsl_obj['old_status'] == ScopeFilterStatus.in_scope:
                self.status_tracker.record_operation(
                    operation=DatabaseOperation.delete,
                    rpsl_pk=rpsl_obj['rpsl_pk'],
                    source=rpsl_obj['source'],
                    object_class=rpsl_obj['object_class'],
                    object_text=rpsl_obj['object_text'],
                    origin=JournalEntryOrigin.scope_filter,
                    source_serial=None,
                )
                self._object_classes_modified.add(rpsl_obj['object_class'])

    def delete_rpsl_object(self, origin: JournalEntryOrigin, rpsl_object: Optional[RPSLObject]=None,
                           source: Optional[str]=None, rpsl_pk: Optional[str]=None,
                           source_serial: Optional[int]=None) -> None:
        """
        Delete an RPSL object from the database.

        The origin indicates the origin of this change, see JournalEntryOrigin
        for the various options.
        """
        self._check_write_permitted()
        self._flush_rpsl_object_writing_buffer()
        table = RPSLDatabaseObject.__table__
        if not source and rpsl_object:
            source = rpsl_object.parsed_data['source']
        if not rpsl_pk and rpsl_object:
            rpsl_pk = rpsl_object.pk()
        stmt = table.delete(
            sa.and_(table.c.rpsl_pk == rpsl_pk, table.c.source == source),
        ).returning(table.c.pk, table.c.rpsl_pk, table.c.source, table.c.object_class, table.c.object_text)
        results = self._connection.execute(stmt)

        if results.rowcount == 0:
            logger.error(f'Attempted to remove object {rpsl_pk}/{source}, but no database row matched')
            return None
        if results.rowcount > 1:  # pragma: no cover
            # This should not be possible, as rpsl_pk/source are a composite unique value in the database scheme.
            # Therefore, a query should not be able to affect more than one row - and we also can not test this
            # scenario. Due to the possible harm of a bug in this area, we still check for it anyways.
            affected_pks = ','.join([r[0] for r in results.fetchall()])
            msg = f'Attempted to remove object {rpsl_pk}/{source}, but multiple objects were affected, '
            msg += f'internal pks affected: {affected_pks}'
            logger.critical(msg)
            raise ValueError(msg)

        result = results.fetchone()
        self.status_tracker.record_operation(
            operation=DatabaseOperation.delete,
            rpsl_pk=result['rpsl_pk'],
            source=result['source'],
            object_class=result['object_class'],
            object_text=result['object_text'],
            origin=origin,
            source_serial=source_serial,
        )
        self._object_classes_modified.add(result['object_class'])

    def _flush_rpsl_object_writing_buffer(self) -> None:
        """
        Flush the current object writing buffer to the database.

        This happens in one large INSERT .. ON CONFLICT DO UPDATE ..
        statement, which is more performant than individual
        queries in case of large datasets.
        """
        if not self._rpsl_upsert_buffer:
            return

        self._check_write_permitted()

        rpsl_composite_key = ['rpsl_pk', 'source']
        stmt = pg.insert(RPSLDatabaseObject).values([x[0] for x in self._rpsl_upsert_buffer])

        if not self._rpsl_guaranteed_no_existing:
            columns_to_update = {
                c.name: c
                for c in stmt.excluded
                if c.name not in rpsl_composite_key and c.name != 'pk'
            }

            stmt = stmt.on_conflict_do_update(
                index_elements=rpsl_composite_key,
                set_=columns_to_update,
            )

        try:
            self._connection.execute(stmt)
        except Exception as exc:  # pragma: no cover
            self._transaction.rollback()
            logger.error(f'Exception occurred while executing statement: {stmt}, rolling back', exc_info=exc)
            raise

        for obj, origin, source_serial in self._rpsl_upsert_buffer:
            # RPKI invalid objects never generated a journal entry,
            # as mirrors should not see them. This does not affect
            # RPKI excluded sources, because they are always not_found.
            # Same applies for scope filter.
            rpki_acceptable = obj['rpki_status'] != RPKIStatus.invalid
            scope_acceptable = obj['scopefilter_status'] == ScopeFilterStatus.in_scope

            if rpki_acceptable and scope_acceptable:
                self.status_tracker.record_operation(
                    operation=DatabaseOperation.add_or_update,
                    rpsl_pk=obj['rpsl_pk'],
                    source=obj['source'],
                    object_class=obj['object_class'],
                    object_text=obj['object_text'],
                    origin=origin,
                    source_serial=source_serial,
                )

        self._rpsl_pk_source_seen = set()
        self._rpsl_upsert_buffer = []

    def _flush_roa_writing_buffer(self):
        """
        Flush the current ROA buffer to the database.
        As ROAs are only ever inserted, never updated,
        this is done with a single COPY command.
        """
        if not self._roa_insert_buffer:
            return

        self._check_write_permitted()

        columns = list(self._roa_insert_buffer[0].keys())
        roa_rows = []
        for roa in self._roa_insert_buffer:
            # ROA data is very simple, so we can use a naive CSV generator.
            roa_data = ','.join([str(roa[k]) for k in columns])
            roa_rows.append(roa_data)

        roa_csv = StringIO('\n'.join(roa_rows))
        roa_csv.seek(0)
        postgres_copy.copy_from(roa_csv, ROADatabaseObject, self._connection, columns=columns, format='csv')
        self._roa_insert_buffer = []

    def delete_all_rpsl_objects_with_journal(self, source, journal_guaranteed_empty=False):
        """
        Delete all RPSL objects for a source from the database,
        all journal entries and the database status.
        This is intended for cases where a full re-import is done.
        Note that no journal records are kept of this change itself.
        """
        self._check_write_permitted()
        self._flush_rpsl_object_writing_buffer()
        table = RPSLDatabaseObject.__table__
        stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        if not journal_guaranteed_empty:
            table = RPSLDatabaseJournal.__table__
            stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        table = RPSLDatabaseStatus.__table__
        stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        # All objects are presumed to have been changed.
        self._object_classes_modified.update(OBJECT_CLASS_MAPPING.keys())

    def delete_all_roa_objects(self):
        """
        Delete all ROA objects from the database.
        ROAs are always imported in bulk.
        """
        self._check_write_permitted()
        self._roa_insert_buffer = []
        stmt = ROADatabaseObject.__table__.delete()
        self._connection.execute(stmt)

    def set_force_reload(self, source):
        """
        Set the force_reload flag for a source.
        Upon the next mirror update, all data for the source will be
        discarded and reloaded from the source.
        """
        self._check_write_permitted()
        table = RPSLDatabaseStatus.__table__
        synchronised_serials = is_serial_synchronised(self, source, settings_only=True)
        stmt = table.update().where(table.c.source == source).values(
            force_reload=True,
            synchronised_serials=synchronised_serials,
            serial_oldest_seen=None,
            serial_newest_seen=None,
        )
        self._connection.execute(stmt)
        logger.info(f'force_reload flag set for {source}, serial synchronisation based on '
                    f'current settings: {synchronised_serials}')

    def record_serial_newest_mirror(self, source: str, serial: int) -> None:
        """
        Record that a mirror was updated to a certain serial.
        """
        self._check_write_permitted()
        self.status_tracker.record_serial_newest_mirror(source, serial)

    def record_serial_seen(self, source: str, serial: int):
        """
        Record that a serial was seen for a source
        """
        self._check_write_permitted()
        self.status_tracker.record_serial_seen(source, serial)

    def record_mirror_error(self, source: str, error: str) -> None:
        """
        Record an error seen in a mirrored database.
        Only the most recent error is stored in the DB status.
        """
        self._check_write_permitted()
        self.status_tracker.record_mirror_error(source, error)

    def record_serial_exported(self, source: str, serial: int) -> None:
        """
        Record an export of a source at a particular serial.
        """
        self._check_write_permitted()
        self.status_tracker.record_serial_exported(source, serial)

    def close(self) -> None:
        self._connection.close()

    def _check_write_permitted(self):
        if self.readonly:
            msg = 'Attempted to write to SQL database from readonly database handler'
            logger.critical(msg)
            raise Exception(msg)


class DatabaseStatusTracker:
    """
    Keep track of the status of sources, and their journal, if enabled.
    Changes to the database should always call record_operation() on this object,
    and finalise_transaction() before committing.

    If journaling is enabled, a new entry in the journal will be made.
    If a journal entry was made, a record is kept in
    memory with all serials encountered/created for this source.

    When finalising, the RPSLDatabaseStatus table is updated to correctly
    reflect the range of serials known for a particular source.
    """
    journaling_enabled: bool
    _new_serials_per_source: Dict[str, Set[int]]
    _sources_seen: Set[str]
    _newest_mirror_serials: Dict[str, int]
    _mirroring_error: Dict[str, str]
    _exported_serials: Dict[str, int]

    c_journal = RPSLDatabaseJournal.__table__.c
    c_status = RPSLDatabaseStatus.__table__.c

    def __init__(self, database_handler: DatabaseHandler, journaling_enabled=True) -> None:
        self.database_handler = database_handler
        self.journaling_enabled = journaling_enabled
        self.reset()

    def record_serial_newest_mirror(self, source: str, serial: int):
        """
        Record that a mirror was updated to a certain serial.
        """
        self._sources_seen.add(source)
        self._newest_mirror_serials[source] = serial

    def record_serial_seen(self, source: str, serial: int):
        """
        Record that a serial was seen for a source
        """
        self._sources_seen.add(source)
        self._new_serials_per_source[source].add(serial)

    def record_mirror_error(self, source: str, error: str) -> None:
        """
        Record an error seen in a mirrored database.
        Only the most recent error is stored in the DB status.
        """
        self._sources_seen.add(source)
        self._mirroring_error[source] = error

    def record_serial_exported(self, source: str, serial: int) -> None:
        """
        Record an export of a source at a particular serial.
        Only the most recent serial is stored in the DB status.
        """
        self._sources_seen.add(source)
        self._exported_serials[source] = serial

    def record_operation(self, operation: DatabaseOperation, rpsl_pk: str, source: str, object_class: str,
                         object_text: str, origin: JournalEntryOrigin, source_serial: Optional[int]) -> None:
        """
        Make a record in the journal of a change to an object.

        Will only record changes when self.journaling_enabled is set,
        and the database.SOURCE.keep_journal is set.
        The source will always be added to _sources_seen.

        Note that this method locks the journal table for writing to ensure a
        gapless set of NRTM serials.
        """
        self._sources_seen.add(source)
        if self.journaling_enabled and get_setting(f'sources.{source}.keep_journal'):
            serial_nrtm: Union[int, sa.sql.expression.Select]
            journal_tablename = RPSLDatabaseJournal.__tablename__

            if self._is_serial_synchronised(source):
                serial_nrtm = source_serial
            else:
                self.database_handler.execute_statement(f'LOCK TABLE {journal_tablename} IN EXCLUSIVE MODE')
                serial_nrtm = sa.select([sa.text('COALESCE(MAX(serial_nrtm), 0) + 1')])
                serial_nrtm = serial_nrtm.where(RPSLDatabaseJournal.__table__.c.source == source)
                serial_nrtm = serial_nrtm.as_scalar()
            stmt = RPSLDatabaseJournal.__table__.insert().values(
                rpsl_pk=rpsl_pk,
                source=source,
                operation=operation,
                object_class=object_class,
                object_text=object_text,
                serial_nrtm=serial_nrtm,
                origin=origin,
            ).returning(self.c_journal.serial_nrtm)

            insert_result = self.database_handler.execute_statement(stmt)
            inserted_serial = insert_result.fetchone()['serial_nrtm']
            self._new_serials_per_source[source].add(inserted_serial)

    def finalise_transaction(self):
        """
        - Create a new status object for all seen sources if it does not exist.
        - Reset the force_reload flag.
        - If new serials were recorded for a source, update the database
          serial stats in the status object.
        - Update the latest source errors.
        """
        for source in self._sources_seen:
            stmt = pg.insert(RPSLDatabaseStatus).values(
                source=source,
                force_reload=False,
                synchronised_serials=self._is_serial_synchronised(source),
                updated=datetime.now(timezone.utc),
            )
            stmt = stmt.on_conflict_do_update(
                index_elements=['source'],
                set_={
                    'force_reload': False,
                    'synchronised_serials': self._is_serial_synchronised(source),
                },
            )
            self.database_handler.execute_statement(stmt)

        for source, serials in self._new_serials_per_source.items():
            serial_oldest_journal_q = sa.select([
                sa.func.min(self.c_journal.serial_nrtm)
            ]).where(self.c_journal.source == source)
            result = self.database_handler.execute_statement(serial_oldest_journal_q)
            serial_oldest_journal = next(result)[0]

            serial_newest_journal_q = sa.select([
                sa.func.max(self.c_journal.serial_nrtm)
            ]).where(self.c_journal.source == source)
            result = self.database_handler.execute_statement(serial_newest_journal_q)
            serial_newest_journal = next(result)[0]

            serial_oldest_seen = sa.select([
                sa.func.least(
                    sa.func.min(self.c_status.serial_oldest_seen),
                    serial_oldest_journal,
                    min(serials)
                )
            ]).where(self.c_status.source == source)
            serial_newest_seen = sa.select([
                sa.func.greatest(
                    sa.func.max(self.c_status.serial_newest_seen),
                    serial_newest_journal,
                    max(serials)
                ),
            ]).where(self.c_status.source == source)

            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                serial_oldest_seen=serial_oldest_seen,
                serial_newest_seen=serial_newest_seen,
                serial_oldest_journal=serial_oldest_journal,
                serial_newest_journal=serial_newest_journal,
                updated=datetime.now(timezone.utc),
            )
            self.database_handler.execute_statement(stmt)

        for source, error in self._mirroring_error.items():
            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                last_error=error,
                last_error_timestamp=datetime.now(timezone.utc),
                updated=datetime.now(timezone.utc),
            )
            self.database_handler.execute_statement(stmt)

        for source, serial in self._newest_mirror_serials.items():
            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                serial_newest_mirror=serial,
                updated=datetime.now(timezone.utc),
            )
            self.database_handler.execute_statement(stmt)

        for source, serial in self._exported_serials.items():
            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                serial_last_export=serial,
            )
            self.database_handler.execute_statement(stmt)

        self.reset()

    @lru_cache(maxsize=100)
    def _is_serial_synchronised(self, source: str) -> bool:
        # Cached wrapper method
        return is_serial_synchronised(self.database_handler, source)

    def reset(self):
        self._new_serials_per_source = defaultdict(set)
        self._sources_seen = set()
        self._newest_mirror_serials = dict()
        self._mirroring_error = dict()
        self._exported_serials = dict()
        self._is_serial_synchronised.cache_clear()


def is_serial_synchronised(database_handler: DatabaseHandler, source: str, settings_only=False) -> bool:
    """
    Determine whether a source should use / is using serial synchronisation
    from the NRTM mirror source.
    If settings_only is set, only look at whether serial synchronisation should
    be enabled based on the current config. Otherwise, also looks at the current
    flag in the database, which catches cases where serials have gone out of
    sync in the past.
    """
    if settings_only:
        db_status = True
    else:
        db_query = DatabaseStatusQuery().source(source)
        db_result = database_handler.execute_query(db_query, flush_rpsl_buffer=False)
        try:
            db_status = next(db_result)['synchronised_serials']
        except StopIteration:
            db_status = True
    settings_status = all([
        not get_setting('scopefilter') or get_setting(f'sources.{source}.scopefilter_excluded'),
        not get_setting('rpki.roa_source') or get_setting(f'sources.{source}.rpki_excluded'),
        get_setting(f'sources.{source}.nrtm_host')
    ])
    return db_status and settings_status



from enum import unique, Enum


@unique
class UpdateRequestType(Enum):
    CREATE = 'create'
    MODIFY = 'modify'
    DELETE = 'delete'


@unique
class UpdateRequestStatus(Enum):
    SAVED = 'saved'
    PROCESSING = 'processing'
    ERROR_UNKNOWN_CLASS = 'error: unknown RPSL class'
    ERROR_PARSING = 'errors encountered during object parsing'
    ERROR_AUTH = 'error: update not authorised'
    ERROR_REFERENCE = 'error: reference to object that does not exist'
    ERROR_ROA = 'error: conflict with existing ROA'
    ERROR_SCOPEFILTER = 'error: not in scope'
    ERROR_NON_AUTHORITIVE = 'error: attempt to update object in non-authoritive database'



import difflib
import logging
from typing import List, Optional, Set

from irrd.conf import get_setting
from irrd.rpki.status import RPKIStatus
from irrd.rpki.validators import SingleRouteROAValidator
from irrd.rpsl.parser import UnknownRPSLObjectClassException, RPSLObject
from irrd.rpsl.rpsl_objects import rpsl_object_from_text, RPSLMntner
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.scopefilter.validators import ScopeFilterValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.models import JournalEntryOrigin
from irrd.storage.queries import RPSLDatabaseQuery
from irrd.utils.text import splitline_unicodesafe
from .parser_state import UpdateRequestType, UpdateRequestStatus
from .validators import ReferenceValidator, AuthValidator

logger = logging.getLogger(__name__)


class ChangeRequest:
    """
    A ChangeRequest tracks and processes a request for a single change.
    In this context, a change can be creating, modifying or deleting an
    RPSL object.
    """
    rpsl_text_submitted: str
    rpsl_obj_new: Optional[RPSLObject]
    rpsl_obj_current: Optional[RPSLObject] = None
    status = UpdateRequestStatus.PROCESSING
    request_type: Optional[UpdateRequestType] = None
    mntners_notify: List[RPSLMntner]

    error_messages: List[str]
    info_messages: List[str]

    def __init__(self, rpsl_text_submitted: str, database_handler: DatabaseHandler, auth_validator: AuthValidator,
                 reference_validator: ReferenceValidator, delete_reason=Optional[str]) -> None:
        """
        Initialise a new change request for a single RPSL object.

        :param rpsl_text_submitted: the object text
        :param database_handler: a DatabaseHandler instance
        :param auth_validator: a AuthValidator instance, to resolve authentication requirements
        :param reference_validator: a ReferenceValidator instance, to resolve references between objects
        :param delete_reason: a string with the deletion reason, if this was a deletion request

        The rpsl_text passed into this function should be cleaned from any
        meta attributes like delete/override/password. Those should be passed
        into this method as delete_reason, or provided to the AuthValidator.

        The auth_validator and reference_validator must be shared between
        different instances, to benefit from caching, and to resolve references
        between different objects that are part of the same submission with
        possibly multiple changes.
        """
        self.database_handler = database_handler
        self.auth_validator = auth_validator
        self.reference_validator = reference_validator
        self.rpsl_text_submitted = rpsl_text_submitted
        self.mntners_notify = []
        self.used_override = False
        self._cached_roa_validity: Optional[bool] = None
        self.roa_validator = SingleRouteROAValidator(database_handler)
        self.scopefilter_validator = ScopeFilterValidator()

        try:
            self.rpsl_obj_new = rpsl_object_from_text(rpsl_text_submitted, strict_validation=True)
            if self.rpsl_obj_new.messages.errors():
                self.status = UpdateRequestStatus.ERROR_PARSING
            self.error_messages = self.rpsl_obj_new.messages.errors()
            self.info_messages = self.rpsl_obj_new.messages.infos()
            logger.debug(f'{id(self)}: Processing new ChangeRequest for object {self.rpsl_obj_new}: request {id(self)}')

        except UnknownRPSLObjectClassException as exc:
            self.rpsl_obj_new = None
            self.request_type = None
            self.status = UpdateRequestStatus.ERROR_UNKNOWN_CLASS
            self.info_messages = []
            self.error_messages = [str(exc)]

        if self.is_valid() and self.rpsl_obj_new:
            source = self.rpsl_obj_new.source()
            if not get_setting(f'sources.{source}.authoritative'):
                logger.debug(f'{id(self)}: change is for non-authoritative source {source}, rejected')
                self.error_messages.append(f'This instance is not authoritative for source {source}')
                self.status = UpdateRequestStatus.ERROR_NON_AUTHORITIVE
                return

            self._retrieve_existing_version()

        if delete_reason:
            self.request_type = UpdateRequestType.DELETE
            if not self.rpsl_obj_current:
                self.status = UpdateRequestStatus.ERROR_PARSING
                self.error_messages.append('Can not delete object: no object found for this key in this database.')
                logger.debug(f'{id(self)}: Request attempts to delete object {self.rpsl_obj_new}, '
                             f'but no existing object found.')

    def _retrieve_existing_version(self):
        """
        Retrieve the current version of this object, if any, and store it in rpsl_obj_current.
        Update self.status appropriately.
        """
        query = RPSLDatabaseQuery().sources([self.rpsl_obj_new.source()])
        query = query.object_classes([self.rpsl_obj_new.rpsl_object_class]).rpsl_pk(self.rpsl_obj_new.pk())
        results = list(self.database_handler.execute_query(query))

        if not results:
            self.request_type = UpdateRequestType.CREATE
            logger.debug(f'{id(self)}: Did not find existing version for object {self.rpsl_obj_new}, request is CREATE')
        elif len(results) == 1:
            self.request_type = UpdateRequestType.MODIFY
            self.rpsl_obj_current = rpsl_object_from_text(results[0]['object_text'], strict_validation=False)
            logger.debug(f'{id(self)}: Retrieved existing version for object '
                         f'{self.rpsl_obj_current}, request is MODIFY/DELETE')
        else:  # pragma: no cover
            # This should not be possible, as rpsl_pk/source are a composite unique value in the database scheme.
            # Therefore, a query should not be able to affect more than one row.
            affected_pks = ', '.join([r['pk'] for r in results])
            msg = f'{id(self)}: Attempted to retrieve current version of object {self.rpsl_obj_new.pk()}/'
            msg += f'{self.rpsl_obj_new.source()}, but multiple '
            msg += f'objects were found, internal pks found: {affected_pks}'
            logger.error(msg)
            raise ValueError(msg)

    def save(self, database_handler: DatabaseHandler) -> None:
        """Save the change to the database."""
        if self.status != UpdateRequestStatus.PROCESSING or not self.rpsl_obj_new:
            raise ValueError('ChangeRequest can only be saved in status PROCESSING')
        if self.request_type == UpdateRequestType.DELETE and self.rpsl_obj_current is not None:
            logger.info(f'{id(self)}: Saving change for {self.rpsl_obj_new}: deleting current object')
            database_handler.delete_rpsl_object(rpsl_object=self.rpsl_obj_current, origin=JournalEntryOrigin.auth_change)
        else:
            logger.info(f'{id(self)}: Saving change for {self.rpsl_obj_new}: inserting/updating current object')
            database_handler.upsert_rpsl_object(self.rpsl_obj_new, JournalEntryOrigin.auth_change)
        self.status = UpdateRequestStatus.SAVED

    def is_valid(self):
        return self.status in [UpdateRequestStatus.SAVED, UpdateRequestStatus.PROCESSING]

    def submitter_report(self) -> str:
        """Produce a string suitable for reporting back status and messages to the human submitter."""
        status = 'succeeded' if self.is_valid() else 'FAILED'

        report = f'{self.request_type_str().title()} {status}: [{self.object_class_str()}] {self.object_pk_str()}\n'
        if self.info_messages or self.error_messages:
            if not self.rpsl_obj_new or self.error_messages:
                report += '\n' + self.rpsl_text_submitted + '\n'
            else:
                report += '\n' + self.rpsl_obj_new.render_rpsl_text() + '\n'
            report += ''.join([f'ERROR: {e}\n' for e in self.error_messages])
            report += ''.join([f'INFO: {e}\n' for e in self.info_messages])
        return report

    def notification_target_report(self):
        """
        Produce a string suitable for reporting back status and messages
        to a human notification target, i.e. someone listed
        in notify/upd-to/mnt-nfy.
        """
        if not self.is_valid() and self.status != UpdateRequestStatus.ERROR_AUTH:
            raise ValueError('Notification reports can only be made for changes that are valid '
                             'or have failed authorisation.')

        status = 'succeeded' if self.is_valid() else 'FAILED AUTHORISATION'
        report = f'{self.request_type_str().title()} {status} for object below: '
        report += f'[{self.object_class_str()}] {self.object_pk_str()}:\n\n'

        if self.request_type == UpdateRequestType.MODIFY:
            current_text = list(splitline_unicodesafe(self.rpsl_obj_current.render_rpsl_text()))
            new_text = list(splitline_unicodesafe(self.rpsl_obj_new.render_rpsl_text()))
            diff = list(difflib.unified_diff(current_text, new_text, lineterm=''))

            report += '\n'.join(diff[2:])  # skip the lines from the diff which would have filenames
            if self.status == UpdateRequestStatus.ERROR_AUTH:
                report += '\n\n*Rejected* new version of this object:\n\n'
            else:
                report += '\n\nNew version of this object:\n\n'

        if self.request_type == UpdateRequestType.DELETE:
            report += self.rpsl_obj_current.render_rpsl_text()
        else:
            report += self.rpsl_obj_new.render_rpsl_text()
        return report

    def request_type_str(self) -> str:
        return self.request_type.value if self.request_type else 'request'

    def object_pk_str(self) -> str:
        return self.rpsl_obj_new.pk() if self.rpsl_obj_new else '(unreadable object key)'

    def object_class_str(self) -> str:
        return self.rpsl_obj_new.rpsl_object_class if self.rpsl_obj_new else '(unreadable object class)'

    def notification_targets(self) -> Set[str]:
        """
        Produce a set of e-mail addresses that should be notified
        about the change to this object.
        May include mntner upd-to or mnt-nfy, and notify of existing object.
        """
        targets: Set[str] = set()
        status_qualifies_notification = self.is_valid() or self.status == UpdateRequestStatus.ERROR_AUTH
        if self.used_override or not status_qualifies_notification:
            return targets

        mntner_attr = 'upd-to' if self.status == UpdateRequestStatus.ERROR_AUTH else 'mnt-nfy'
        for mntner in self.mntners_notify:
            for email in mntner.parsed_data.get(mntner_attr, []):
                targets.add(email)

        if self.rpsl_obj_current:
            for email in self.rpsl_obj_current.parsed_data.get('notify', []):
                targets.add(email)

        return targets

    def validate(self) -> bool:
        auth_valid = self._check_auth()
        if not auth_valid:
            return False
        references_valid = self._check_references()
        rpki_valid = self._check_conflicting_roa()
        scopefilter_valid = self._check_scopefilter()
        return references_valid and rpki_valid and scopefilter_valid

    def _check_auth(self) -> bool:
        assert self.rpsl_obj_new
        auth_result = self.auth_validator.process_auth(self.rpsl_obj_new, self.rpsl_obj_current)
        self.info_messages += auth_result.info_messages
        self.mntners_notify = auth_result.mntners_notify

        if not auth_result.is_valid():
            self.status = UpdateRequestStatus.ERROR_AUTH
            self.error_messages += auth_result.error_messages
            logger.debug(f'{id(self)}: Authentication check failed: {auth_result.error_messages}')
            return False

        self.used_override = auth_result.used_override

        logger.debug(f'{id(self)}: Authentication check succeeded')
        return True

    def _check_references(self) -> bool:
        """
        Check all references from this object to or from other objects.

        For deletions, only references to the deleted object matter, as
        they now become invalid. For other operations, only the validity
        of references from the new object to others matter.
        """
        if self.request_type == UpdateRequestType.DELETE and self.rpsl_obj_current is not None:
            assert self.rpsl_obj_new
            references_result = self.reference_validator.check_references_from_others(self.rpsl_obj_current)
        else:
            assert self.rpsl_obj_new
            references_result = self.reference_validator.check_references_to_others(self.rpsl_obj_new)
        self.info_messages += references_result.info_messages

        if not references_result.is_valid():
            self.error_messages += references_result.error_messages
            logger.debug(f'{id(self)}: Reference check failed: {references_result.error_messages}')
            if self.is_valid():  # Only change the status if this object was valid prior, so this is the first failure
                self.status = UpdateRequestStatus.ERROR_REFERENCE
            return False

        logger.debug(f'{id(self)}: Reference check succeeded')
        return True

    def _check_conflicting_roa(self) -> bool:
        """
        Check whether there are any conflicting ROAs with the new object.
        Result is cached, as validate() may be called multiple times,
        but the result of this check will not change. Always returns
        True when not in RPKI-aware mode.
        """
        assert self.rpsl_obj_new
        if self._cached_roa_validity is not None:
            return self._cached_roa_validity
        if not get_setting('rpki.roa_source') or not self.rpsl_obj_new.rpki_relevant:
            return True
        # Deletes are permitted for RPKI-invalids, other operations are not
        if self.request_type == UpdateRequestType.DELETE:
            return True

        assert self.rpsl_obj_new.asn_first
        validation_result = self.roa_validator.validate_route(
            self.rpsl_obj_new.prefix, self.rpsl_obj_new.asn_first, self.rpsl_obj_new.source()
        )
        if validation_result == RPKIStatus.invalid:
            import_timer = get_setting('rpki.roa_import_timer')
            user_message = 'RPKI ROAs were found that conflict with this object. '
            user_message += f'(This IRRd refreshes ROAs every {import_timer} seconds.)'
            logger.debug(f'{id(self)}: Conflicting ROAs found')
            if self.is_valid():  # Only change the status if this object was valid prior, so this is first failure
                self.status = UpdateRequestStatus.ERROR_ROA
            self.error_messages.append(user_message)
            self._cached_roa_validity = False
            return False
        else:
            logger.debug(f'{id(self)}: No conflicting ROAs found')
        self._cached_roa_validity = True
        return True

    def _check_scopefilter(self) -> bool:
        if self.request_type == UpdateRequestType.DELETE or not self.rpsl_obj_new:
            return True
        result, comment = self.scopefilter_validator.validate_rpsl_object(self.rpsl_obj_new)
        if result in [ScopeFilterStatus.out_scope_prefix, ScopeFilterStatus.out_scope_as]:
            user_message = 'Contains out of scope information: ' + comment
            if self.request_type == UpdateRequestType.CREATE:
                logger.debug(f'{id(self)}: object out of scope: ' + comment)
                if self.is_valid():  # Only change the status if this object was valid prior, so this is first failure
                    self.status = UpdateRequestStatus.ERROR_SCOPEFILTER
                self.error_messages.append(user_message)
                return False
            elif self.request_type == UpdateRequestType.MODIFY:
                self.info_messages.append(user_message)
        return True


def parse_change_requests(requests_text: str,
                          database_handler: DatabaseHandler,
                          auth_validator: AuthValidator,
                          reference_validator: ReferenceValidator,
                          ) -> List[ChangeRequest]:
    """
    Parse change requests, a text of RPSL objects along with metadata like
    passwords or deletion requests.

    :param requests_text: a string containing all change requests
    :param database_handler: a DatabaseHandler instance
        :param auth_validator: a AuthValidator instance, to resolve authentication requirements
    :param reference_validator: a ReferenceValidator instance
    :return: a list of ChangeRequest instances
    """
    results = []
    passwords = []
    overrides = []

    requests_text = requests_text.replace('\r', '')
    for object_text in requests_text.split('\n\n'):
        object_text = object_text.strip()
        if not object_text:
            continue

        rpsl_text = ''
        delete_reason = None

        # The attributes password/override/delete are meta attributes
        # and need to be extracted before parsing. Delete refers to a specific
        # object, password/override apply to all included objects.
        for line in splitline_unicodesafe(object_text):
            if line.startswith('password:'):
                password = line.split(':', maxsplit=1)[1].strip()
                passwords.append(password)
            elif line.startswith('override:'):
                override = line.split(':', maxsplit=1)[1].strip()
                overrides.append(override)
            elif line.startswith('delete:'):
                delete_reason = line.split(':', maxsplit=1)[1].strip()
            else:
                rpsl_text += line + '\n'

        if not rpsl_text:
            continue

        results.append(ChangeRequest(rpsl_text, database_handler, auth_validator, reference_validator,
                                     delete_reason=delete_reason))

    if auth_validator:
        auth_validator.passwords = passwords
        auth_validator.overrides = overrides
    return results



"""initial db

Revision ID: 28dc1cd85bdc
Revises:
Create Date: 2018-06-11 14:37:13.472465

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '28dc1cd85bdc'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # Creating this extension requires extra permissions.
    # However, if it is already created, this command succeeds
    # even if this user does not have sufficient permissions.
    op.execute('CREATE EXTENSION IF NOT EXISTS "pgcrypto";')

    op.create_table('database_status',
                    sa.Column('pk', postgresql.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'),
                              nullable=False),
                    sa.Column('source', sa.String(), nullable=False),
                    sa.Column('serial_oldest_seen', sa.Integer(), nullable=True),
                    sa.Column('serial_newest_seen', sa.Integer(), nullable=True),
                    sa.Column('serial_oldest_journal', sa.Integer(), nullable=True),
                    sa.Column('serial_newest_journal', sa.Integer(), nullable=True),
                    sa.Column('serial_last_export', sa.Integer(), nullable=True),
                    sa.Column('force_reload', sa.Boolean(), nullable=False),
                    sa.Column('last_error', sa.Text(), nullable=True),
                    sa.Column('last_error_timestamp', sa.DateTime(timezone=True), nullable=True),
                    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.Column('updated', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.PrimaryKeyConstraint('pk')
                    )
    op.create_index(op.f('ix_database_status_source'), 'database_status', ['source'], unique=True)

    op.create_table('rpsl_database_journal',
                    sa.Column('pk', postgresql.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'),
                              nullable=False),
                    sa.Column('rpsl_pk', sa.String(), nullable=False),
                    sa.Column('source', sa.String(), nullable=False),
                    sa.Column('serial_nrtm', sa.Integer(), nullable=False),
                    sa.Column('operation', sa.Enum('add_or_update', 'delete', name='databaseoperation'),
                              nullable=False),
                    sa.Column('object_class', sa.String(), nullable=False),
                    sa.Column('object_text', sa.Text(), nullable=False),
                    sa.Column('timestamp', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.PrimaryKeyConstraint('pk'),
                    sa.UniqueConstraint('serial_nrtm', 'source', name='rpsl_objects_history_serial_nrtm_source_unique')
                    )
    op.create_index(op.f('ix_rpsl_database_journal_object_class'), 'rpsl_database_journal', ['object_class'],
                    unique=False)
    op.create_index(op.f('ix_rpsl_database_journal_rpsl_pk'), 'rpsl_database_journal', ['rpsl_pk'], unique=False)
    op.create_index(op.f('ix_rpsl_database_journal_serial_nrtm'), 'rpsl_database_journal', ['serial_nrtm'],
                    unique=False)
    op.create_index(op.f('ix_rpsl_database_journal_source'), 'rpsl_database_journal', ['source'], unique=False)

    op.create_table('rpsl_objects',
                    sa.Column('pk', postgresql.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'),
                              nullable=False),
                    sa.Column('rpsl_pk', sa.String(), nullable=False),
                    sa.Column('source', sa.String(), nullable=False),
                    sa.Column('object_class', sa.String(), nullable=False),
                    sa.Column('parsed_data', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
                    sa.Column('object_text', sa.Text(), nullable=False),
                    sa.Column('ip_version', sa.Integer(), nullable=True),
                    sa.Column('ip_first', postgresql.INET(), nullable=True),
                    sa.Column('ip_last', postgresql.INET(), nullable=True),
                    sa.Column('ip_size', sa.DECIMAL(scale=0), nullable=True),
                    sa.Column('asn_first', sa.BigInteger(), nullable=True),
                    sa.Column('asn_last', sa.BigInteger(), nullable=True),
                    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.Column('updated', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.PrimaryKeyConstraint('pk'),
                    sa.UniqueConstraint('rpsl_pk', 'source', name='rpsl_objects_rpsl_pk_source_unique')
                    )
    op.create_index(op.f('ix_rpsl_objects_ip_first'), 'rpsl_objects', ['ip_first'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_ip_last'), 'rpsl_objects', ['ip_last'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_asn_first'), 'rpsl_objects', ['asn_first'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_asn_last'), 'rpsl_objects', ['asn_last'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_ip_version'), 'rpsl_objects', ['ip_version'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_object_class'), 'rpsl_objects', ['object_class'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_rpsl_pk'), 'rpsl_objects', ['rpsl_pk'], unique=False)
    op.create_index(op.f('ix_rpsl_objects_source'), 'rpsl_objects', ['source'], unique=False)
    op.create_index('ix_rpsl_objects_asn_first_asn_last', 'rpsl_objects', ['asn_first', 'asn_last'], unique=False)
    op.create_index('ix_rpsl_objects_ip_first_ip_last', 'rpsl_objects', ['ip_first', 'ip_last'], unique=False)
    op.create_index('ix_rpsl_objects_ip_last_ip_first', 'rpsl_objects', ['ip_last', 'ip_first'], unique=False)

    # Manually added
    op.create_index(op.f('ix_rpsl_objects_parsed_data_admin_c'), 'rpsl_objects',
                    [sa.text("((parsed_data->'admin-c'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_tech_c'), 'rpsl_objects', [sa.text("((parsed_data->'tech-c'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_zone_c'), 'rpsl_objects', [sa.text("((parsed_data->'zone-c'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_role'), 'rpsl_objects', [sa.text("((parsed_data->'role'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_members'), 'rpsl_objects',
                    [sa.text("((parsed_data->'members'))")], unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_person'), 'rpsl_objects', [sa.text("((parsed_data->'person'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_mnt_by'), 'rpsl_objects', [sa.text("((parsed_data->'mnt-by'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_member_of'), 'rpsl_objects',
                    [sa.text("((parsed_data->'member-of'))")], unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_mp_members'), 'rpsl_objects',
                    [sa.text("((parsed_data->'mp-members'))")], unique=False,
                    postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_origin'), 'rpsl_objects', [sa.text("((parsed_data->'origin'))")],
                    unique=False, postgresql_using='gin')
    op.create_index(op.f('ix_rpsl_objects_parsed_data_mbrs_by_ref'), 'rpsl_objects',
                    [sa.text("((parsed_data->'mbrs-by-ref'))")],
                    unique=False, postgresql_using='gin')


def downgrade():
    # Manually added
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_mbrs_by_ref'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_ip_first'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_ip_last'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_zone_c'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_role'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_members'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_person'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_mnt_by'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_member_of'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_parsed_data_mp_members'), table_name='rpsl_objects')

    # Autogenerated
    op.drop_index('ix_rpsl_objects_ip_last_ip_first', table_name='rpsl_objects')
    op.drop_index('ix_rpsl_objects_ip_first_ip_last', table_name='rpsl_objects')
    op.drop_index('ix_rpsl_objects_asn_first_asn_last', table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_source'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_rpsl_pk'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_object_class'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_ip_version'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_asn_last'), table_name='rpsl_objects')
    op.drop_index(op.f('ix_rpsl_objects_asn_first'), table_name='rpsl_objects')
    op.drop_table('rpsl_objects')

    op.drop_index(op.f('ix_rpsl_database_journal_source'), table_name='rpsl_database_journal')
    op.drop_index(op.f('ix_rpsl_database_journal_serial_nrtm'), table_name='rpsl_database_journal')
    op.drop_index(op.f('ix_rpsl_database_journal_rpsl_pk'), table_name='rpsl_database_journal')
    op.drop_index(op.f('ix_rpsl_database_journal_object_class'), table_name='rpsl_database_journal')
    op.drop_table('rpsl_database_journal')

    op.drop_index(op.f('ix_database_status_source'), table_name='database_status')
    op.drop_table('database_status')



"""Add prefix_length and rpki_status to rpsl_objects

Revision ID: 64a3d6faf6d4
Revises: e07863eac52f
Create Date: 2019-03-04 14:39:14.711874

"""
# revision identifiers, used by Alembic.
import sqlalchemy as sa
from alembic import op

revision = '64a3d6faf6d4'
down_revision = 'e07863eac52f'
branch_labels = None
depends_on = None


def upgrade():
    rpki_status = sa.Enum('valid', 'invalid', 'not_found', name='rpkistatus')
    rpki_status.create(op.get_bind())

    op.add_column('rpsl_objects', sa.Column('rpki_status', sa.Enum('valid', 'invalid', 'not_found', name='rpkistatus'), nullable=False, server_default='not_found'))
    op.create_index(op.f('ix_rpsl_objects_rpki_status'), 'rpsl_objects', ['rpki_status'], unique=False)

    op.add_column('rpsl_objects', sa.Column('prefix_length', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('rpsl_objects', 'prefix_length')
    op.drop_index(op.f('ix_rpsl_objects_rpki_status'), table_name='rpsl_objects')
    op.drop_column('rpsl_objects', 'rpki_status')

    rpki_status = sa.Enum('valid', 'invalid', 'not_found', name='rpkistatus')
    rpki_status.drop(op.get_bind())



"""bogon_to_scopefilter

Revision ID: 4a514ead8fc2
Revises: 39e4f15ed80c
Create Date: 2020-07-09 20:11:45.873381

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.exc import ProgrammingError

revision = '4a514ead8fc2'
down_revision = '39e4f15ed80c'
branch_labels = None
depends_on = None


def upgrade():
    scopefilter_status = sa.Enum('in_scope', 'out_scope_as', 'out_scope_prefix', name='scopefilterstatus')
    scopefilter_status.create(op.get_bind())

    op.add_column('rpsl_objects', sa.Column('scopefilter_status', sa.Enum('in_scope', 'out_scope_as', 'out_scope_prefix', name='scopefilterstatus'), server_default='in_scope', nullable=False))
    op.create_index(op.f('ix_rpsl_objects_scopefilter_status'), 'rpsl_objects', ['scopefilter_status'], unique=False)

    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    columns = inspector.get_columns('rpsl_objects')

    # This is a somewhat strange migration: the bogon_status column may or may
    # not exist, depending on which version of 39e4f15ed80c people have run.
    # If it does exist, it is deleted.
    # This is also why downgrade() does not recreate this column.
    if 'bogon_status' in columns:
        op.drop_index('ix_rpsl_objects_bogon_status', table_name='rpsl_objects')
        op.drop_column('rpsl_objects', 'bogon_status')
        bogon_status = sa.Enum('unknown', 'not_bogon', 'bogon_as', 'bogon_prefix', name='bogonstatus')
        bogon_status.drop(op.get_bind())

    # downgrade() can't remove this entry from the enum, so if this migration
    # is reverted and then re-applied, altering the enum will fail
    with op.get_context().autocommit_block():
        try:
            op.execute("ALTER TYPE journalentryorigin ADD VALUE 'scope_filter'")
        except ProgrammingError as pe:
            if 'DuplicateObject' not in str(pe):
                raise pe


def downgrade():
    op.drop_index(op.f('ix_rpsl_objects_scopefilter_status'), table_name='rpsl_objects')
    op.drop_column('rpsl_objects', 'scopefilter_status')

    scopefilter_status = sa.Enum('in_scope', 'out_scope_as', 'out_scope_prefix', name='scopefilterstatus')
    scopefilter_status.drop(op.get_bind())



"""Add roa_object table.

Revision ID: e07863eac52f
Revises: 28dc1cd85bdc
Create Date: 2019-02-28 16:03:57.797697

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'e07863eac52f'
down_revision = '28dc1cd85bdc'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('roa_object',
                    sa.Column('pk', postgresql.UUID(as_uuid=True), server_default=sa.text('gen_random_uuid()'),
                              nullable=False),
                    sa.Column('prefix', postgresql.CIDR(), nullable=False),
                    sa.Column('asn', sa.BigInteger(), nullable=False),
                    sa.Column('max_length', sa.Integer(), nullable=False),
                    sa.Column('trust_anchor', sa.String(), nullable=True),
                    sa.Column('ip_version', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(timezone=True), server_default=sa.text('now()'), nullable=False),
                    sa.PrimaryKeyConstraint('pk'),
                    sa.UniqueConstraint('prefix', 'asn', 'max_length', 'trust_anchor', name='roa_object_prefix_asn_maxlength_unique')
                    )
    op.create_index(op.f('ix_roa_object_ip_version'), 'roa_object', ['ip_version'], unique=False)
    op.create_index('ix_roa_objects_prefix_gist', 'roa_object', [sa.text('prefix inet_ops')], unique=False,
                    postgresql_using='gist')


def downgrade():
    op.drop_index(op.f('ix_roa_objects_prefix_gist'), table_name='roa_object')
    op.drop_index(op.f('ix_roa_object_ip_version'), table_name='roa_object')
    op.drop_table('roa_object')



import logging
import socket
import textwrap
from typing import Optional

from beautifultable import BeautifulTable

from irrd import __version__
from irrd.conf import get_setting
from irrd.conf.defaults import DEFAULT_SOURCE_NRTM_PORT
from irrd.storage.database_handler import DatabaseHandler, is_serial_synchronised
from irrd.storage.queries import DatabaseStatusQuery, RPSLDatabaseObjectStatisticsQuery
from irrd.utils.whois_client import whois_query_source_status

logger = logging.getLogger(__name__)


class StatusGenerator:

    def generate_status(self) -> str:
        """
        Generate a human-readable overview of database status.
        """
        database_handler = DatabaseHandler()

        statistics_query = RPSLDatabaseObjectStatisticsQuery()
        self.statistics_results = list(database_handler.execute_query(statistics_query))
        status_query = DatabaseStatusQuery()
        self.status_results = list(database_handler.execute_query(status_query))

        results = [
            self._generate_header(),
            self._generate_statistics_table(),
            self._generate_source_detail(database_handler)
        ]
        database_handler.close()
        return '\n\n'.join(results)

    def _generate_header(self) -> str:
        """
        Generate the header of the report, containing basic info like version
        and time until the next mirror update.
        """
        return textwrap.dedent(f"""
        IRRD version {__version__}
        Listening on {get_setting('server.whois.interface')} port {get_setting('server.whois.port')}
        """).lstrip()

    def _generate_statistics_table(self) -> str:
        """
        Generate a table with an overview of basic stats for each database.
        """
        table = BeautifulTable(default_alignment=BeautifulTable.ALIGN_RIGHT)
        table.column_headers = ['source', 'total obj', 'rt obj', 'aut-num obj', 'serial', 'last export']
        table.column_alignments['source'] = BeautifulTable.ALIGN_LEFT
        table.left_border_char = table.right_border_char = ''
        table.right_border_char = table.bottom_border_char = ''
        table.row_separator_char = ''
        table.column_separator_char = '  '

        for status_result in self.status_results:
            source = status_result['source'].upper()
            total_obj, route_obj, autnum_obj = self._statistics_for_source(source)
            serial = status_result['serial_newest_seen']
            last_export = status_result['serial_last_export']
            if not last_export:
                last_export = ''
            table.append_row([source, total_obj, route_obj, autnum_obj, serial, last_export])

        total_obj, route_obj, autnum_obj = self._statistics_for_source(None)
        table.append_row(['TOTAL', total_obj, route_obj, autnum_obj, '', ''])

        return str(table)

    def _statistics_for_source(self, source: Optional[str]):
        """
        Extract counts of total objects, route objects and aut-num objects,
        from the results of a previous SQL query.
        If source is None, all sources are counted.
        """
        if source:
            source_statistics = [s for s in self.statistics_results if s['source'] == source]
        else:
            source_statistics = self.statistics_results

        total_obj = sum([s['count'] for s in source_statistics])
        route_obj = sum([s['count'] for s in source_statistics if s['object_class'] == 'route'])
        autnum_obj = sum([s['count'] for s in source_statistics if s['object_class'] == 'aut-num'])
        return total_obj, route_obj, autnum_obj

    def _generate_source_detail(self, database_handler: DatabaseHandler) -> str:
        """
        Generate status details for each database.

        This includes local configuration, local database status metadata,
        and serial information queried from the remote NRTM host,
        queried by _generate_remote_status_info().
        :param database_handler:
        """
        result_txt = ''
        for status_result in self.status_results:
            source = status_result['source'].upper()
            keep_journal = 'Yes' if get_setting(f'sources.{source}.keep_journal') else 'No'
            authoritative = 'Yes' if get_setting(f'sources.{source}.authoritative') else 'No'
            object_class_filter = get_setting(f'sources.{source}.object_class_filter')
            rpki_enabled = get_setting('rpki.roa_source') and not get_setting(f'sources.{source}.rpki_excluded')
            rpki_enabled_str = 'Yes' if rpki_enabled else 'No'
            scopefilter_enabled = get_setting('rpki.scopefiler') and not get_setting(f'sources.{source}.scopefilter_excluded')
            scopefilter_enabled_str = 'Yes' if scopefilter_enabled else 'No'
            synchronised_serials_str = 'Yes' if is_serial_synchronised(database_handler, source) else 'No'

            nrtm_host = get_setting(f'sources.{source}.nrtm_host')
            nrtm_port = int(get_setting(f'sources.{source}.nrtm_port', DEFAULT_SOURCE_NRTM_PORT))

            remote_information = self._generate_remote_status_info(nrtm_host, nrtm_port, source)
            remote_information = textwrap.indent(remote_information, ' ' * 16)

            result_txt += textwrap.dedent(f"""
            Status for {source}
            -------------------
            Local information:
                Authoritative: {authoritative}
                Object class filter: {object_class_filter}
                Oldest serial seen: {status_result['serial_oldest_seen']}
                Newest serial seen: {status_result['serial_newest_seen']}
                Oldest local journal serial number: {status_result['serial_oldest_journal']}
                Newest local journal serial number: {status_result['serial_newest_journal']}
                Last export at serial number: {status_result['serial_last_export']}
                Newest serial number mirrored: {status_result['serial_newest_mirror']}
                Synchronised NRTM serials: {synchronised_serials_str}
                Last update: {status_result['updated']}
                Local journal kept: {keep_journal}
                Last import error occurred at: {status_result['last_error_timestamp']}
                RPKI validation enabled: {rpki_enabled_str}
                Scope filter enabled: {scopefilter_enabled_str}

            Remote information:{remote_information}
            """)
        return result_txt

    def _generate_remote_status_info(self, nrtm_host: Optional[str], nrtm_port: int, source: str) -> str:
        """
        Determine the remote status.

        If NRTM is configured, this will include querying the NRTM
        source for serial information. Various error states will produce
        an appropriate remote status message for the report.
        """
        if nrtm_host:
            try:
                source_status = whois_query_source_status(nrtm_host, nrtm_port, source)
                mirrorable, mirror_serial_oldest, mirror_serial_newest, mirror_export_serial = source_status
                mirrorable_str = 'Yes' if mirrorable else 'No'

                return textwrap.dedent(f"""
                    NRTM host: {nrtm_host} port {nrtm_port}
                    Mirrorable: {mirrorable_str}
                    Oldest journal serial number: {mirror_serial_oldest}
                    Newest journal serial number: {mirror_serial_newest}
                    Last export at serial number: {mirror_export_serial}
                    """)
            except ValueError:
                return textwrap.dedent(f"""
                    NRTM host: {nrtm_host} port {nrtm_port}
                    Remote status query unsupported or query failed
                    """)
            except (socket.timeout, ConnectionError):
                return textwrap.dedent(f"""
                    NRTM host: {nrtm_host} port {nrtm_port}
                    Unable to reach remote server for status query
                    """)
        else:
            return textwrap.dedent("""
                No NRTM host configured.
                """)



import logging
import signal

from IPy import IP
from setproctitle import setproctitle
import socket

import socketserver
from http import HTTPStatus
# http.server is generally not recommended for production use,
# but as we only use small parts, this is not a significant concern
from http.server import HTTPServer, BaseHTTPRequestHandler
from typing import Tuple

from irrd import __version__
from irrd.conf import get_setting
from irrd.server.access_check import is_client_permitted
from irrd.server.http.status_generator import StatusGenerator

logger = logging.getLogger(__name__)
HTTP_TIMEOUT = 30

# The start_http_server method and IRRdHTTPRequestHandler are
# difficult to test in a unit test, and therefore only included
# in integration tests. They should be kept as small as possible.


def start_http_server():  # pragma: no cover
    """
    Start the HTTP server, listening forever. This function does not return.
    """
    setproctitle('irrd-http-server-listener')
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    address = (get_setting('server.http.interface'), get_setting('server.http.port'))
    logger.info(f'Starting http server on TCP {address}')
    httpd = HTTPServerForkingIPv6(address, IRRdHTTPRequestHandler)
    httpd.serve_forever()


class HTTPServerForkingIPv6(socketserver.ThreadingMixIn, HTTPServer):  # pragma: no cover
    # Default HTTP server only supports IPv4
    allow_reuse_address = True
    timeout = HTTP_TIMEOUT

    def __init__(self, server_address, RequestHandlerClass, bind_and_activate=True):  # noqa: N803
        self.address_family = socket.AF_INET6 if IP(server_address[0]).version() == 6 else socket.AF_INET
        super().__init__(server_address, RequestHandlerClass, bind_and_activate)

    def handle_error(self, request, client_address):
        logger.error(f'Error while handling request from {client_address}', exc_info=True)


class IRRdHTTPRequestHandler(BaseHTTPRequestHandler):  # pragma: no cover
    """
    Request handler, called for each HTTP request, in its own process.

    Kept as lightweight as possible, and offloads most work to
    IRRdHTTPRequestProcessor, as that class is unit-testable.
    """
    server_version = f'irrd/{__version__}'

    def handle_one_request(self) -> None:
        # Sockets for client connections don't inherit the timeout,
        # this wrapper sets it before any data is read.
        self.connection.settimeout(HTTP_TIMEOUT)
        return super().handle_one_request()

    def do_GET(self):  # noqa: N802
        processor = IRRdHTTPRequestProcessor(self.client_address[0], self.client_address[1])
        status, content = processor.handle_get(self.path)
        self.generate_response(status, content)

    def log_message(self, format, *args):
        logger.info('HTTP request served: %s: %s' % (self.address_string(), format % args))

    def generate_response(self, response_status, content):
        content_bytes = content.encode('utf-8')
        self.send_response(response_status)
        self.send_header('Content-type', 'text/plain;charset=utf-8')
        self.send_header('Content-Length', len(content_bytes))
        self.end_headers()
        self.wfile.write(content_bytes)


class IRRdHTTPRequestProcessor:
    """
    Request processor for HTTP queries.
    Split from IRRdHTTPRequestHandler for easier testing.
    """
    def __init__(self, client_ip: str, client_port: int):
        self.client_ip = client_ip
        self.client_str = client_ip + ':' + str(client_port)

    def handle_get(self, path) -> Tuple[HTTPStatus, str]:
        if not is_client_permitted(self.client_ip, 'server.http.access_list'):
            return HTTPStatus.FORBIDDEN, 'Access denied'

        if path.rstrip('/') != '/v1/status':
            return HTTPStatus.NOT_FOUND, 'Not found'

        content = StatusGenerator().generate_status()
        return HTTPStatus.OK, content



import re
from typing import Iterator, Union, TextIO, Optional

from irrd.conf import PASSWORD_HASH_DUMMY_VALUE
from irrd.rpsl.config import PASSWORD_HASHERS

re_remove_passwords = re.compile(r'(%s)[^\n]+' % '|'.join(PASSWORD_HASHERS.keys()), flags=re.IGNORECASE)
re_remove_last_modified = re.compile(r'^last-modified: [^\n]+\n', flags=re.MULTILINE)


def remove_auth_hashes(input: Optional[str]):
    if not input:
        return input
    return re_remove_passwords.sub(r'\1 %s  # Filtered for security' % PASSWORD_HASH_DUMMY_VALUE, input)


def remove_last_modified(rpsl_text: str):
    """
    Remove all last-modified attributes from an RPSL text with less overhead
    than using the full RPSL parser.
    Assumes the last-modified value is single line. This is a safe assumption
    when the input is guaranteed to have been generated by IRRd.
    """
    return re_remove_last_modified.sub('', rpsl_text)


def splitline_unicodesafe(input: str) -> Iterator[str]:
    """
    Split an input string by newlines, and return an iterator of the lines.

    This is a replacement for Python's built-in splitlines, which also splits
    on characters such as unicode line separator (U+2028). In RPSL, that should
    not be considered a line separator.
    """
    if not input:
        return
    for line in input.strip('\n').split('\n'):
        yield line.strip('\r')


def split_paragraphs_rpsl(input: Union[str, TextIO], strip_comments=True) -> Iterator[str]:
    """
    Split an input into paragraphs, and return an iterator of the paragraphs.

    A paragraph is a block of text, separated by at least one empty line.
    Note that a line with other whitespace, e.g. a space, is not considered
    empty.

    If strip_comments=True, any line starting with % or # is entirely ignored,
    both within a paragraph and between paragraphs.
    """
    current_paragraph = ''
    if isinstance(input, str):
        generator = splitline_unicodesafe(input)
    else:
        generator = input

    for line in generator:
        line = line.strip('\r\n')
        if strip_comments and line.startswith('%') or line.startswith('#'):
            continue
        if line:
            current_paragraph += line + '\n'
        if not line:
            if current_paragraph:
                yield current_paragraph
            current_paragraph = ''

    if current_paragraph.strip():
        yield current_paragraph



import logging
from typing import List, Optional

import sqlalchemy as sa
from IPy import IP
from sqlalchemy.sql import Select, ColumnCollection

from irrd.rpki.status import RPKIStatus
from irrd.rpsl.rpsl_objects import lookup_field_names
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.storage.models import (RPSLDatabaseObject, RPSLDatabaseJournal, RPSLDatabaseStatus,
                                 ROADatabaseObject)
from irrd.utils.validators import parse_as_number, ValidationError

logger = logging.getLogger(__name__)


class BaseRPSLObjectDatabaseQuery:
    statement: Select
    table: sa.Table
    columns: ColumnCollection

    def __init__(self, ordered_by_sources=True, enable_ordering=True):
        self._query_frozen = False
        self._sources_list = []
        self._ordered_by_sources = ordered_by_sources
        self._enable_ordering = enable_ordering

    def pk(self, pk: str):
        """Filter on an exact object PK (UUID)."""
        return self._filter(self.columns.pk == pk)

    def rpsl_pk(self, rpsl_pk: str):
        """Filter on an exact RPSL PK (e.g. 192.0.2.0/24,AS65537)."""
        return self.rpsl_pks([rpsl_pk])

    def rpsl_pks(self, rpsl_pks: List[str]):
        """Filter on an exact RPSL PK (e.g. 192.0.2.0/24,AS65537) - will match any PK in the list."""
        rpsl_pks = [p.upper().strip() for p in rpsl_pks]
        return self._filter(self.columns.rpsl_pk.in_(rpsl_pks))

    def sources(self, sources: List[str]):
        """
        Filter on one or more sources.

        Sources list must be an iterable. Will match objects from any
        of the mentioned sources. Order is used for sorting of results.
        """
        sources = [s.upper().strip() for s in sources]
        self._sources_list = sources
        fltr = self.columns.source.in_(self._sources_list)
        return self._filter(fltr)

    def object_classes(self, object_classes: List[str]):
        """
        Filter on one or more object classes.

        Classes list must be an iterable. Will match objects from any
        of the mentioned classes.
        """
        fltr = self.columns.object_class.in_(object_classes)
        return self._filter(fltr)

    def first_only(self):
        """Only return the first match."""
        self.statement = self.statement.limit(1)
        return self

    def finalise_statement(self) -> Select:
        """
        Finalise the statement and return it.

        This method does some final work on statements that may be dependent on
        each other - particularly statements that determine the sort order of
        the query, which depends on sources_list() and prioritise_source().
        """
        self._query_frozen = True

        if self._enable_ordering:
            order_by = []
            if 'ip_first' in self.columns:
                order_by.append(self.columns.ip_first.asc())
            if 'asn_first' in self.columns:
                order_by.append(self.columns.asn_first.asc())
            if 'rpsl_pk' in self.columns:
                order_by.append(self.columns.rpsl_pk.asc())

            if self._ordered_by_sources and self._sources_list:
                case_elements = []
                for idx, source in enumerate(self._sources_list):
                    case_elements.append((self.columns.source == source, idx + 1))

                criterion = sa.case(case_elements, else_=100000)
                order_by.insert(0, criterion)

            self.statement = self.statement.order_by(*order_by)
        return self.statement

    def _filter(self, fltr):
        self._check_query_frozen()
        self.statement = self.statement.where(fltr)
        return self

    def _check_query_frozen(self) -> None:
        if self._query_frozen:
            raise ValueError('This query was frozen - no more filters can be applied.')


class RPSLDatabaseQuery(BaseRPSLObjectDatabaseQuery):
    """
    RPSL data query builder for retrieving RPSL objects.

    Offers various ways to filter, which are always constructed in an AND query.
    For example:
        q = RPSLDatabaseQuery().sources(['NTTCOM']).asn_less_specific(65537)
    would match all objects that refer or include AS65537 (i.e. aut-num, route,
    as-block, route6) from the NTTCOM source.

    For methods taking a prefix or IP address, this should be an IPy.IP object.
    """
    table = RPSLDatabaseObject.__table__
    columns = RPSLDatabaseObject.__table__.c
    lookup_field_names = lookup_field_names()

    def __init__(self, column_names=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if column_names is None:
            columns = [
                self.columns.pk,
                self.columns.object_class,
                self.columns.rpsl_pk,
                self.columns.parsed_data,
                self.columns.object_text,
                self.columns.source,
                self.columns.rpki_status,
            ]
        else:
            columns = [self.columns.get(name) for name in column_names]
        self.statement = sa.select(columns)
        self._lookup_attr_counter = 0

    def lookup_attr(self, attr_name: str, attr_value: str):
        """
        Filter on a lookup attribute, e.g. mnt-by.
        At least one of the values for the lookup attribute must match attr_value.
        Matching is case-insensitive.
        """
        return self.lookup_attrs_in([attr_name], [attr_value])

    def lookup_attrs_in(self, attr_names: List[str], attr_values: List[str]):
        """
        Filter on one or more lookup attributes, e.g. mnt-by, or ['admin-c', 'tech-c']
        At least one of the values for at least one of the lookup attributes must
        match one of the items in attr_values. Matching is case-insensitive.
        """
        attr_names = [attr_name.lower() for attr_name in attr_names]
        for attr_name in attr_names:
            if attr_name not in self.lookup_field_names:
                raise ValueError(f'Invalid lookup attribute: {attr_name}')
        self._check_query_frozen()

        value_filters = []
        statement_params = {}
        for attr_name in attr_names:
            for attr_value in attr_values:
                counter = self._lookup_attr_counter
                self._lookup_attr_counter += 1
                value_filters.append(sa.text(f'parsed_data->:lookup_attr_name{counter} ? :lookup_attr_value{counter}'))
                statement_params[f'lookup_attr_name{counter}'] = attr_name
                statement_params[f'lookup_attr_value{counter}'] = attr_value.upper()
        fltr = sa.or_(*value_filters)
        self.statement = self.statement.where(fltr).params(**statement_params)

        return self

    def ip_exact(self, ip: IP):
        """
        Filter on an exact prefix or address.

        The provided ip should be an IPy.IP class, and can be a prefix or
        an address.
        """
        fltr = sa.and_(
            self.columns.ip_first == str(ip.net()),
            self.columns.ip_last == str(ip.broadcast()),
            self.columns.ip_version == ip.version()
        )
        return self._filter(fltr)

    def ip_less_specific(self, ip: IP):
        """Filter any less specifics or exact matches of a prefix."""
        fltr = sa.and_(
            self.columns.ip_first <= str(ip.net()),
            self.columns.ip_last >= str(ip.broadcast()),
            self.columns.ip_version == ip.version()
        )
        return self._filter(fltr)

    def ip_less_specific_one_level(self, ip: IP):
        """
        Filter one level less specific of a prefix.

        Due to implementation details around filtering, this must
        always be the last call on a query object, or unpredictable
        results may occur.
        """
        self._check_query_frozen()
        # One level less specific could still have multiple objects.
        # A subquery determines the smallest possible size less specific object,
        # and this is then used to filter for any objects with that size.
        fltr = sa.and_(
            self.columns.ip_first <= str(ip.net()),
            self.columns.ip_last >= str(ip.broadcast()),
            self.columns.ip_version == ip.version(),
            sa.not_(sa.and_(self.columns.ip_first == str(ip.net()), self.columns.ip_last == str(ip.broadcast()))),
        )
        self.statement = self.statement.where(fltr)

        size_subquery = self.statement.with_only_columns([self.columns.ip_size])
        size_subquery = size_subquery.order_by(self.columns.ip_size.asc())
        size_subquery = size_subquery.limit(1)

        self.statement = self.statement.where(self.columns.ip_size.in_(size_subquery))
        self._query_frozen = True
        return self

    def ip_more_specific(self, ip: IP):
        """Filter any more specifics of a prefix, not including exact matches.

        Note that this only finds full more specifics: objects for which their
        IP range is fully encompassed by the ip parameter.
        """
        fltr = sa.and_(
            self.columns.ip_first >= str(ip.net()),
            self.columns.ip_first <= str(ip.broadcast()),
            self.columns.ip_last <= str(ip.broadcast()),
            self.columns.ip_last >= str(ip.net()),
            self.columns.ip_version == ip.version(),
            sa.not_(sa.and_(self.columns.ip_first == str(ip.net()), self.columns.ip_last == str(ip.broadcast()))),
        )
        return self._filter(fltr)

    def asn(self, asn: int):
        """
        Filter for exact matches on an ASN.
        """
        fltr = sa.and_(self.columns.asn_first == asn, self.columns.asn_last == asn)
        return self._filter(fltr)

    def asns_first(self, asns: List[int]):
        """
        Filter for asn_first being in a list of ASNs.
        This is useful when also restricting object class to 'route' for instance.
        """
        fltr = self.columns.asn_first.in_(asns)
        return self._filter(fltr)

    def asn_less_specific(self, asn: int):
        """
        Filter for a specific ASN, or any less specific matches.

        This will match all objects that refer to this ASN, or a block
        encompassing it - including route, route6, aut-num and as-block.
        """
        fltr = sa.and_(self.columns.asn_first <= asn, self.columns.asn_last >= asn)
        return self._filter(fltr)

    def rpki_status(self, status: List[RPKIStatus]):
        """
        Filter for RPSL objects with a specific RPKI validation status.
        """
        fltr = self.columns.rpki_status.in_(status)
        return self._filter(fltr)

    def scopefilter_status(self, status: List[ScopeFilterStatus]):
        """
        Filter for RPSL objects with a specific scope filter status.
        """
        fltr = self.columns.scopefilter_status.in_(status)
        return self._filter(fltr)

    def text_search(self, value: str):
        """
        Search the database for a specific free text.

        In order, this attempts:
        - If the value is a valid AS number, return all as-block, as-set, aut-num objects
          relating or including that AS number.
        - If the value is a valid IP address or network, return all objects that relate to
          that resource and any less specifics.
        - Otherwise, return all objects where the RPSL primary key is exactly this value,
          or it matches part of a person/role name (not nic-hdl, their
          actual person/role attribute value).
        """
        self._check_query_frozen()
        try:
            _, asn = parse_as_number(value)
            return self.object_classes(['as-block', 'as-set', 'aut-num']).asn_less_specific(asn)
        except ValidationError:
            pass

        try:
            ip = IP(value)
            return self.ip_less_specific(ip)
        except ValueError:
            pass

        counter = self._lookup_attr_counter
        self._lookup_attr_counter += 1
        fltr = sa.or_(
            self.columns.rpsl_pk == value.upper(),
            sa.and_(
                self.columns.object_class == 'person',
                sa.text(f"parsed_data->>'person' ILIKE :lookup_attr_text_search{counter}")
            ),
            sa.and_(
                self.columns.object_class == 'role',
                sa.text(f"parsed_data->>'role' ILIKE :lookup_attr_text_search{counter}")
            ),
        )
        self.statement = self.statement.where(fltr).params(
            **{f'lookup_attr_text_search{counter}': '%' + value + '%'}
        )
        return self

    def __repr__(self):
        return f'RPSLDatabaseQuery: {self.statement}\nPARAMS: {self.statement.compile().params}'


class RPSLDatabaseJournalQuery(BaseRPSLObjectDatabaseQuery):
    """
    RPSL data query builder for retrieving the journal,
    analogous to RPSLDatabaseQuery.
    """
    table = RPSLDatabaseJournal.__table__
    columns = RPSLDatabaseJournal.__table__.c

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statement = sa.select([
            self.columns.pk,
            self.columns.rpsl_pk,
            self.columns.source,
            self.columns.serial_nrtm,
            self.columns.operation,
            self.columns.object_class,
            self.columns.object_text,
            self.columns.timestamp,
        ]).order_by(self.columns.source.asc(), self.columns.serial_nrtm.asc())

    def serial_range(self, start: int, end: Optional[int]=None):
        """
        Filter for a serials within a specific range, inclusive.
        """
        if end is not None:
            fltr = sa.and_(self.columns.serial_nrtm >= start, self.columns.serial_nrtm <= end)
        else:
            fltr = self.columns.serial_nrtm >= start
        return self._filter(fltr)

    def __repr__(self):
        return f'RPSLDatabaseJournalQuery: {self.statement}\nPARAMS: {self.statement.compile().params}'


class DatabaseStatusQuery:
    table = RPSLDatabaseStatus.__table__
    columns = RPSLDatabaseStatus.__table__.c

    def __init__(self):
        self._sources_list: List[str] = []
        self.statement = sa.select([
            self.columns.pk,
            self.columns.source,
            self.columns.serial_oldest_seen,
            self.columns.serial_newest_seen,
            self.columns.serial_oldest_journal,
            self.columns.serial_newest_journal,
            self.columns.serial_last_export,
            self.columns.serial_newest_mirror,
            self.columns.force_reload,
            self.columns.synchronised_serials,
            self.columns.last_error,
            self.columns.last_error_timestamp,
            self.columns.created,
            self.columns.updated,
        ])

    def source(self, source: str):
        """Filter on a source."""
        return self.sources([source])

    def sources(self, sources: List[str]):
        """Filter on one or more sources."""
        self._sources_list = [s.upper() for s in sources]
        return self

    def finalise_statement(self):
        order_by = [self.columns.source.asc()]

        if self._sources_list:
            fltr = self.columns.source.in_(self._sources_list)
            self._filter(fltr)

            case_elements = []
            for idx, source in enumerate(self._sources_list):
                case_elements.append((self.columns.source == source, idx + 1))

            criterion = sa.case(case_elements, else_=100000)
            order_by.insert(0, criterion)

        self.statement = self.statement.order_by(*order_by)
        return self.statement

    def _filter(self, fltr):
        self.statement = self.statement.where(fltr)
        return self

    def __repr__(self):
        return f'DatabaseStatusQuery: {self.statement}\nPARAMS: {self.statement.compile().params}'


class RPSLDatabaseObjectStatisticsQuery:
    """
    Special statistics query, calculating the number of
    objects per object class per source.
    """
    table = RPSLDatabaseObject.__table__
    columns = RPSLDatabaseObject.__table__.c

    def __init__(self):
        self.statement = sa.select([
            self.columns.source,
            self.columns.object_class,
            sa.func.count(self.columns.pk).label('count'),
        ]).group_by(self.columns.source, self.columns.object_class)

    def finalise_statement(self):
        return self.statement

    def __repr__(self):
        return f'RPSLDatabaseObjectStatisticsQuery: {self.statement}\nPARAMS: {self.statement.compile().params}'


class ROADatabaseObjectQuery:
    """
    Query builder for ROA objects.
    """
    table = ROADatabaseObject.__table__
    columns = ROADatabaseObject.__table__.c

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statement = sa.select([
            self.columns.pk,
            self.columns.prefix,
            self.columns.asn,
            self.columns.max_length,
            self.columns.trust_anchor,
            self.columns.ip_version,
        ])

    def ip_less_specific_or_exact(self, ip: IP):
        """Filter any less specifics or exact matches of a prefix."""
        fltr = sa.and_(
            self.columns.prefix.op('>>=')(str(ip))
        )
        self.statement = self.statement.where(fltr)
        return self

    def finalise_statement(self):
        return self.statement

    def __repr__(self):
        return f'ROADatabaseObjectQuery: {self.statement}\nPARAMS: {self.statement.compile().params}'
